'''
Created on 22 juil. 2022

@author: tleduc

Copyright 2020-2022 Thomas Leduc

This file is part of t4gpd.

t4gpd is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

t4gpd is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with t4gpd.  If not, see <https://www.gnu.org/licenses/>.
'''
from datetime import datetime
from os.path import splitext

from geopandas.geodataframe import GeoDataFrame
from t4gpd.commons.GeomLib import GeomLib
from t4gpd.commons.IllegalArgumentTypeException import IllegalArgumentTypeException

from t4gpd.io.SalomeWriter import SalomeWriter


class SalomeWriterAndExtruder(SalomeWriter):
    '''
    classdocs
    '''

    def __init__(self, inputGdf, outputFile, elevationFieldname='HAUTEUR',
                 withFaceIds=False, exportBrep=True):
        '''
        Constructor
        '''
        if not isinstance(inputGdf, GeoDataFrame):
            raise IllegalArgumentTypeException(inputGdf, 'GeoDataFrame')
        self.inputGdf = inputGdf

        self.outputFile = splitext(outputFile)[0]

        if elevationFieldname not in inputGdf:
                raise Exception(f'{elevationFieldname} is not a relevant field name!')
        self.elevationFieldname = elevationFieldname

        self.withFaceIds = withFaceIds
        self.exportBrep = exportBrep

    def _dumpBody(self, out):
        cntFaces, cntCtrs, cntNodes = 1, 1, 1
        
        for _, row in self.inputGdf.iterrows():
            geom = row.geometry
            elevation = row[self.elevationFieldname]
            if ('Polygon' == geom.geom_type):
                cntFaces, cntCtrs, cntNodes = self._dumpPolygon(out, geom, elevation, cntFaces, cntCtrs, cntNodes)
            else:
                print(f'SalomeWriter does not handle {geom.geom_type}!')

    def _dumpHeader(self, out):
        out.write(f'''"""
# t4gpd (SalomeWriterAndExtruder), {datetime.now().strftime('%Y-%m-%d %X')}
# ================================================================
# To be copied and pasted into SALOME's console:
with open('{self.outputFile}.py') as f:
    exec(compile(f.read(), '{self.outputFile}.py', 'exec'))
# ================================================================
"""
import sys
sys.stdout.flush()
import salome
salome.salome_init()
import GEOM
from salome.geom import geomBuilder
geompy = geomBuilder.New()
gg = salome.ImportComponentGUI('GEOM')
isPlanarWanted = 1

# ================================================================
''')

    def _dumpPolygon(self, out, geom, elevation, cntFaces, cntCtrs, cntNodes):
        out.write(f'\n# ===== FACE {cntFaces}: CONTOUR {cntCtrs} =====\n')
        cntNode0 = cntNodes
        cntFaces, cntCtrs, cntNodes = self._dumpLinearRing(
            out, geom.exterior, cntFaces, cntCtrs, cntNodes, lbl=str(cntFaces))

        if GeomLib.isHoled(geom):
            nHoles = len(geom.interiors)
            out.write(f'\n# -- Subtract {nHoles} hole(s) --\n')
            for numHole, hole in enumerate(geom.interiors, start=1):
                lbl = f'{cntFaces}_H{numHole}'
                cntFaces, cntCtrs, cntNodes = self._dumpLinearRing(
                    out, hole, cntFaces, cntCtrs, cntNodes, lbl)
                out.write(f'face{cntFaces} = geompy.MakeCut(face{cntFaces}, face{lbl})\n\n')

        if self.withFaceIds:
            out.write(f'''
id_face{cntFaces} = geompy.addToStudy(face{cntFaces}, 'Face_{cntFaces}')
gg.createAndDisplayGO(id_face{cntFaces})
gg.setDisplayMode(id_face{cntFaces}, 1)
''')
        if (1 == cntFaces):
            out.write(f'allBuildFaces = face1\n')
        else:
            out.write(f'allBuildFaces = geompy.MakeFuse(allBuildFaces, face{cntFaces})\n')

        coords = geom.exterior.coords[0]
        x, y, z = coords[0], coords[1], (0 if (2 == len(coords)) else coords[2]) + elevation
        out.write(f'''
ptZ{cntNode0} = geompy.MakeVertex({x}, {y}, {z})
building{cntFaces} = geompy.MakePrism(face{cntFaces}, pt{cntNode0}, ptZ{cntNode0})
id_building{cntFaces} = geompy.addToStudy(building{cntFaces}, 'Building_{cntFaces}')
gg.createAndDisplayGO(id_building{cntFaces})
''')

        cntFaces += 1
        return cntFaces, cntCtrs, cntNodes

'''
from t4gpd.commons.GeomLib import GeomLib
from t4gpd.demos.GeoDataFrameDemos import GeoDataFrameDemos
from t4gpd.io.SalomeWriterAndExtruder import SalomeWriterAndExtruder

gdf = GeoDataFrameDemos.ensaNantesBuildings()
gdf.geometry = gdf.geometry.apply(lambda g: GeomLib.forceZCoordinateToZ0(g, 0))
SalomeWriterAndExtruder(gdf, '/tmp/salome_script2.py', elevationFieldname='HAUTEUR',
                        withFaceIds=False, exportBrep=False).run()
'''



'''
Created on 17 juin 2020

@author: tleduc

Copyright 2020 Thomas Leduc

This file is part of t4gpd.

t4gpd is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

t4gpd is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with t4gpd.  If not, see <https://www.gnu.org/licenses/>.
'''
from geopandas.geodataframe import GeoDataFrame
from numpy import ceil
from shapely.affinity import translate
from shapely.geometry import LineString, Polygon

from t4gpd.commons.BoundingBox import BoundingBox
from t4gpd.commons.GeoProcess import GeoProcess
from t4gpd.commons.GeomLib import GeomLib
from t4gpd.commons.IllegalArgumentTypeException import IllegalArgumentTypeException


class SVGWriter(GeoProcess):
    '''
    classdocs
    '''

    def __init__(self, inputGdf, outputFile, bbox=None, color='black'):
        '''
        Constructor
        '''
        if not isinstance(inputGdf, GeoDataFrame):
            raise IllegalArgumentTypeException(inputGdf, 'GeoDataFrame')
        self.inputGdf = inputGdf

        self.outputFile = outputFile
        self.bbox = BoundingBox(self.inputGdf) if bbox is None else BoundingBox(bbox)
        self.color = color

    def __convertListOfCoords(self, height, coords):
        n = len(coords)
        return ' '.join(["%c %g,%g" % ('M' if (0 == i) else 'L', coords[i][0], height - coords[i][1]) for i in range(n)])

    def __convertGeom(self, height, geom):
        if isinstance(geom, LineString):
            tmp = self.__convertListOfCoords(height, geom.coords)
            return '\t<path d="%s" style="fill:none;stroke:%s;stroke-width:3.0" />\n' % (tmp, self.color)

        elif isinstance(geom, Polygon):
            rings = [geom.exterior] + list(geom.interiors)
            tmp = ''.join([self.__convertListOfCoords(height, ring.coords) + ' z ' for ring in rings])
            return '\t<path d="%s" style="fill:%s;stroke:white;stroke-width:0.025" />\n' % (tmp, self.color)

        return ''

    def run(self):
        x0, y0, h, w = self.bbox.minx, self.bbox.miny, self.bbox.height(), self.bbox.width()

        with open(self.outputFile, 'w') as f: 
            f.write('<?xml version="1.0" encoding="UTF-8" standalone="no"?>\n')
            f.write('<svg height="%d" width="%d">\n' % (ceil(h), ceil(w)))

            for _, row in self.inputGdf.iterrows():
                geom = row.geometry
                geom = translate(geom, -x0, -y0)
                
                if GeomLib.isMultipart(geom):
                    for g in geom.geoms:
                        f.write(self.__convertGeom(h, g))
                else:
                    f.write(self.__convertGeom(h, geom))

            f.write('</svg>\n')



'''
Created on 18 nov. 2020

@author: tleduc

Copyright 2020 Thomas Leduc

This file is part of t4gpd.

t4gpd is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

t4gpd is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with t4gpd.  If not, see <https://www.gnu.org/licenses/>.
'''
from geopandas.geodataframe import GeoDataFrame
from numpy import pi
from shapely.coords import CoordinateSequence
from shapely.geometry import LinearRing, LineString, Polygon
from t4gpd.commons.AngleLib import AngleLib
from t4gpd.commons.GeoProcess import GeoProcess
from t4gpd.commons.GeomLib import GeomLib
from t4gpd.commons.IllegalArgumentTypeException import IllegalArgumentTypeException

import matplotlib.pyplot as plt


class StreetOrientationHistogram(GeoProcess):
    '''
    classdocs
    '''

    def __init__(self, inputGdf, nClusters, outputFile='', title=None, density=False, cumulative=False):
        '''
        Constructor
        '''
        if not isinstance(inputGdf, GeoDataFrame):
            raise IllegalArgumentTypeException(inputGdf, 'GeoDataFrame')
        self.inputGdf = inputGdf

        self.nClusters = nClusters
        self.outputFile = outputFile
        self.title = title
        self.density = density
        self.cumulative = cumulative

    def __fromShapelyGeometryToAzimuthsAndLengths(self, obj):
        result = []

        if GeomLib.isMultipart(obj):
            for g in obj.geoms:
                result += self.__fromShapelyGeometryToAzimuthsAndLengths(g)
            return result

        elif isinstance(obj, (LinearRing, LineString)):
            return self.__fromShapelyGeometryToAzimuthsAndLengths(obj.coords)

        elif isinstance(obj, Polygon):
            result = self.__fromShapelyGeometryToAzimuthsAndLengths(obj.exterior.coords)
            for hole in obj.interiors:
                result += self.__fromShapelyGeometryToAzimuthsAndLengths(hole.coords)
            return result

        elif isinstance(obj, (list, tuple, CoordinateSequence)):
            for i in range(1, len(obj)):
                _azim = AngleLib.angleBetweenNodes((obj[i - 1][0] + 1, obj[i - 1][1]), obj[i - 1], obj[i])
                _azim = AngleLib.toDegrees(_azim if (_azim < pi) else _azim - pi)
                _len = GeomLib.distFromTo(obj[i - 1], obj[i])
                result.append([_azim, _len])

        return result

    def run(self):
        items, totalLen = [], 0.0
        for geom in self.inputGdf.geometry:
            items += self.__fromShapelyGeometryToAzimuthsAndLengths(geom)
            totalLen += geom.length

        x = [_azim for _azim, _ in items]
        weights = [_len for _, _len in items]

        mainFig = plt.figure(1)
        mainFig.clear()

        if self.title is None:
            plt.title('Street orientation histogram\nRotation in the counter clockwise direction starting from full East',
                      fontsize=14)
        else:
            plt.title(self.title, fontsize=14)
        plt.xlabel('Orientation (0 - 180$^0$)')
        if self.density:
            plt.ylabel('Probability density')
        else:
            plt.ylabel('Cumulative road length (m)')
        plt.grid(True)

        n, bins, _ = plt.hist(x, bins=self.nClusters, density=self.density, cumulative=self.cumulative,
                              range=(0, 180),
                              weights=weights,
                              facecolor='gray', edgecolor='black', orientation='vertical',
                              hatch='', histtype='bar', align='mid', rwidth=1.0)

        xx = [(bins[i] + bins[i + 1]) / 2.0 for i in range(len(bins) - 1)]
        plt.plot(xx, n, 'ro', xx, n, 'r--')

        if self.outputFile is not None:
            if ('' == self.outputFile):
                plt.show()
            else:
                plt.savefig(self.outputFile, format='pdf', bbox_inches='tight')  # , facecolor='w', edgecolor='k', pad_inches=0.15)

        return [n, bins]



'''
Created on 12 juil. 2021

@author: tleduc

Copyright 2020-2021 Thomas Leduc

This file is part of t4gpd.

t4gpd is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

t4gpd is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with t4gpd.  If not, see <https://www.gnu.org/licenses/>.
'''
from tempfile import TemporaryDirectory

from geopandas import read_file
from geopandas.geodataframe import GeoDataFrame
from t4gpd.commons.IllegalArgumentTypeException import IllegalArgumentTypeException


class Reloading(object):
    '''
    classdocs
    '''

    def __init__(self, gdf):
        '''
        Constructor
        '''
        if not isinstance(gdf, GeoDataFrame):
            raise IllegalArgumentTypeException(gdf, 'GeoDataFrame')
        self.gdf = gdf

    def run(self):
        with TemporaryDirectory() as tmpdir:
            ofile = f'{tmpdir}/temporary.gpkg'
            self.gdf.to_file(ofile, driver='GPKG')
            return read_file(ofile)
        raise Exception('Unreachable instruction!')



'''
Created on 17 juin 2020

@author: tleduc

Copyright 2020 Thomas Leduc

This file is part of t4gpd.

t4gpd is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

t4gpd is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with t4gpd.  If not, see <https://www.gnu.org/licenses/>.
'''
from shapely.geometry import Point

from t4gpd.commons.Epsilon import Epsilon


class MyNode(object):
    '''
    classdocs
    '''

    @staticmethod
    def isnumeric(x):
        return isinstance(x, (int, float))

    def __init__(self, *args):
        '''
        Constructor
        '''
        if (1 == len(args)) and isinstance(args[0], Point):
            self.x, self.y, self.z = [args[0].x, args[0].y] + [0.0]

        elif (1 == len(args)) and isinstance(args[0], (list, tuple)) and (2 <= len(args[0])):
            self.x, self.y, self.z = list(args[0]) + [0.0] if (2 == len(args[0])) else args[0]

        elif (2 <= len(args)) and MyNode.isnumeric(args[0]) and MyNode.isnumeric(args[1]):
            self.x, self.y = [ args[0], args[1] ]
            self.z = args[2] if (3 <= len(args)) else 0.0

        else:
            raise RuntimeError('Initialization error with %s of type %s' % (args, type(args)))
        self.hash = None

    def foo(self, x):
        # TODO: comprendre pourquoi j'ai besoin de ce patch debile ?!
        return x if isinstance(x, float) else x()

    def __eq__(self, other):
        selfx, selfy, selfz = [ self.foo(self.x), self.foo(self.y), self.foo(self.z)]
        otherx, othery, otherz = [ self.foo(other.x), self.foo(other.y), self.foo(other.z)]
        result = (isinstance(other, MyNode) and Epsilon.equals(selfx, otherx) and 
                  Epsilon.equals(selfy, othery) and Epsilon.equals(selfz, otherz))
        return result 

    def __hash__(self):
        selfx, selfy, selfz = [ self.foo(self.x), self.foo(self.y), self.foo(self.z)]

        if self.hash is None:
            self.hash = 17
            self.hash = 37 * self.hash + hash(selfx)
            self.hash = 37 * self.hash + hash(selfy)
            self.hash = 37 * self.hash + hash(selfz)
        return self.hash

    def __repr__(self):
        selfx, selfy, selfz = [ self.foo(self.x), self.foo(self.y), self.foo(self.z)]
        return '(%.1f, %.1f, %.1f)' % (selfx, selfy, selfz)



'''
Created on 24 juin 2022

@author: tleduc

Copyright 2020-2022 Thomas Leduc

This file is part of t4gpd.

t4gpd is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

t4gpd is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with t4gpd.  If not, see <https://www.gnu.org/licenses/>.
'''
from shapely.geometry import JOIN_STYLE
from t4gpd.morph.geoProcesses.AbstractGeoprocess import AbstractGeoprocess


class RepresentativePoint(AbstractGeoprocess):
    '''
    classdocs
    '''

    def __init__(self, buffdist=4, join_style=JOIN_STYLE.mitre):
        '''
        Constructor
        '''
        self.buffdist = buffdist
        self.join_style = join_style

    def representativePoint(self, geom, buffdist):
        _geom = geom.buffer(-buffdist, join_style=self.join_style)
        if _geom.is_empty:
            return None
        return _geom.representative_point()

    def runWithArgs(self, row):
        geom = row.geometry
        buffdist = self.buffdist
        for _ in range(10):
            p = self.representativePoint(geom, buffdist)
            if p is not None:
                return { 'geometry': p }
            buffdist /= 2
        return { 'geometry': geom.representative_point() }



'''
Created on 12 mai 2021

@author: tleduc

Copyright 2020-2021 Thomas Leduc

This file is part of t4gpd.

t4gpd is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

t4gpd is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with t4gpd.  If not, see <https://www.gnu.org/licenses/>.
'''
from t4gpd.comfort.algo.CommonsLib import CommonsLib
from t4gpd.comfort.algo.ConstantsLib import ConstantsLib
from t4gpd.comfort.algo.VDI_PET_corrected import pet, system


class PETLib(object):
    '''
    classdocs
    '''

    @staticmethod
    def assess_pet(AirTC, RH, WS_ms, T_mrt):
        M_activity = ConstantsLib.M  # [W]
        icl = ConstantsLib.Clo  # [clo]

        # Air pressure in Pascal and mPa
        _, P_air = CommonsLib.airPressure(AirTC, RH)

        Tair, Tmrt, v_air, pvap = AirTC, T_mrt, WS_ms, 10.0 * P_air

        tc, tsk, tcl, esw_real = system(Tair, Tmrt, pvap, v_air, M_activity, icl)
        tsk, enbal, esw, ed, PET, tcl = pet(tc, tsk, tcl, Tair, esw_real)

        return tsk, tc , tcl, PET



'''
Created on 23 juin 2022

@author: tleduc

Copyright 2020-2022 Thomas Leduc

This file is part of t4gpd.

t4gpd is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

t4gpd is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with t4gpd.  If not, see <https://www.gnu.org/licenses/>.
'''
from geopandas import GeoDataFrame
from numpy import sqrt
from shapely.affinity import scale
from t4gpd.commons.GeoProcess import GeoProcess

from t4gpd.pyvista.commons.Triangle3D import Triangle3D


class Icosahedron(GeoProcess):
    '''
    classdocs
    '''
    SQR5 = sqrt(5)
    INVSQR5 = 1.0 / SQR5
    ALPHA = sqrt((3 - SQR5) / 10)
    BETA = sqrt((5 + SQR5) / 10)
    GAMMA = -(1 + SQR5) / (2 * SQR5)
    DELTA = sqrt((5 - SQR5) / 10)

    def __init__(self, norecursions, radius=None):
        '''
        Constructor
        '''
        p = [
            (0, 0, 1),
            (2 * Icosahedron.INVSQR5, 0, Icosahedron.INVSQR5),
            (Icosahedron.ALPHA, Icosahedron.BETA, Icosahedron.INVSQR5),
            (Icosahedron.GAMMA, Icosahedron.DELTA, Icosahedron.INVSQR5),
            (Icosahedron.GAMMA, -Icosahedron.DELTA, Icosahedron.INVSQR5),
            (Icosahedron.ALPHA, -Icosahedron.BETA, Icosahedron.INVSQR5),
            
            (-Icosahedron.ALPHA, Icosahedron.BETA, -Icosahedron.INVSQR5),
            (-Icosahedron.GAMMA, Icosahedron.DELTA, -Icosahedron.INVSQR5),
            (-Icosahedron.GAMMA, -Icosahedron.DELTA, -Icosahedron.INVSQR5),
            (-Icosahedron.ALPHA, -Icosahedron.BETA, -Icosahedron.INVSQR5),
            (-2 * Icosahedron.INVSQR5, 0, -Icosahedron.INVSQR5),
            (0, 0, -1)
            ]
        # build faces
        self.listOfTriangles = [
            Triangle3D(p[0], p[1], p[2]),
            Triangle3D(p[0], p[2], p[3]),
            Triangle3D(p[0], p[3], p[4]),
            Triangle3D(p[0], p[4], p[5]),
            Triangle3D(p[0], p[5], p[1]),
            Triangle3D(p[1], p[7], p[2]),
            Triangle3D(p[2], p[6], p[3]),
            Triangle3D(p[3], p[10], p[4]),
            Triangle3D(p[4], p[9], p[5]),
            Triangle3D(p[5], p[8], p[1]),

            Triangle3D(p[11], p[9], p[10]),
            Triangle3D(p[11], p[8], p[9]),
            Triangle3D(p[11], p[7], p[8]),
            Triangle3D(p[11], p[6], p[7]),
            Triangle3D(p[11], p[10], p[6]),
            Triangle3D(p[10], p[9], p[4]),
            Triangle3D(p[9], p[5], p[8]),
            Triangle3D(p[8], p[1], p[7]),
            Triangle3D(p[7], p[2], p[6]),
            Triangle3D(p[6], p[3], p[10])
            ]
        self.norecursions = norecursions
        self.radius = radius

    @staticmethod
    def fromNRaysToNRecursions(nrays):
        return 0 if (20 >= nrays) else 1 + Icosahedron.fromNRaysToNRecursions(nrays / 4)

    @staticmethod
    def getNRaysOptions():
        return (20, 80, 320, 1280, 5120, 20480, 81920, 327680)

    def getRays(self):
        rows = self.__divide(self.listOfTriangles, self.norecursions)
        return [row.centroid() for row in rows]

    def __divide(self, listOfTriangles, norecursions):
        result = []
        if (0 >= norecursions):
            for tri in listOfTriangles:
                result.append(tri)
        else:
            for tri in listOfTriangles:
                result += self.__divide(tri.divideBy4OnSphere(), norecursions - 1)
        return result

    def run(self):
        rows = self.__divide(self.listOfTriangles, self.norecursions)
        if (self.radius is None) or (1 == self.radius):
            rows = [{'geometry': row.toPolygon()} for row in rows]
        else:
            rows = [{'geometry': scale(
                row.toPolygon(),
                xfact=self.radius,
                yfact=self.radius,
                zfact=self.radius,
                origin=(0, 0, 0))} for row in rows]
        return GeoDataFrame(rows)



'''
Created on 25 aug. 2020

@author: tleduc

Copyright 2020 Thomas Leduc

This file is part of t4gpd.

t4gpd is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

t4gpd is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with t4gpd.  If not, see <https://www.gnu.org/licenses/>.
'''
from numpy import exp, pi, sin
from t4gpd.commons.AngleLib import AngleLib


class PerrinDeBrichambaut(object):
    '''
    classdocs

    Irradiance is understood as instantaneous density of solar radiation incident on a given
    surface, typically expressed in W/m2.

    Irradiation is the sum of irradiance over a time period (e.g. 1 hour, day, month, etc.)
    expressed in J/m2. However, in daily routine Wh/m2 are more commonly used.
    '''
    PURE_SKY = 0
    STANDARD_SKY = 1
    POLLUTED_SKY = 2

    @staticmethod
    def diffuseSolarIrradiance(solarAltitudeAngle, skyType=STANDARD_SKY):
        '''
        Ghodbane, M., & Boumeddane, B. (2016). Estimating solar radiation according to semi 
        empirical approach of Perrin de Brichambaut: application on several areas with different 
        climate in Algeria. International Journal of Energetica, 1(1), 20. 
        https://doi.org/10.47238/ijeca.v1i1.12
        '''
        assert (0 <= solarAltitudeAngle <= (pi / 2)), 'solarAltitudeAngle in radians!'

        if (PerrinDeBrichambaut.PURE_SKY == skyType):
            D = 0.75
        elif (PerrinDeBrichambaut.STANDARD_SKY == skyType):
            D = 1
        elif (PerrinDeBrichambaut.POLLUTED_SKY == skyType):
            D = 4 / 3

        return 125 * D * sin(solarAltitudeAngle) ** 0.4

    @staticmethod
    def directNormalIrradiance(solarAltitudeAngle, skyType=STANDARD_SKY):
        '''
        Direct Normal Irradiance is the amount of solar radiation received per unit area by a
        surface that is always held perpendicular (or normal) to the rays that come in a
        straight line from the direction of the sun at its current position in the sky.

        d'apres p.105 de :
        AFEDES. 1979. Memosol : Memento D'heliotechnique. ed. Editions Europeennes Thermique 
        et Industrie (EETI). Paris, France.

        et (Miguet, 2000; p. 168) :
        Eclairement energetique (W/m2) solaire direct et normal (pour une surface 
        perpendiculaire aux rayons solaires), en conditions d'insolation normales
        '''
        assert (0 <= solarAltitudeAngle <= (pi / 2)), 'solarAltitudeAngle in radians!'

        angle = AngleLib.toDegrees(solarAltitudeAngle)

        if (PerrinDeBrichambaut.PURE_SKY == skyType):
            A, B, C = 1210, 6, 1
        elif (PerrinDeBrichambaut.STANDARD_SKY == skyType):
            A, B, C = 1230, 4, 2
            # A, B, C = 1230, 3.8, 1.6
        elif (PerrinDeBrichambaut.POLLUTED_SKY == skyType):
            A, B, C = 1260, 2.3, 3

        return A * exp(-1 / (B * sin(AngleLib.toRadians(angle + C))))



'''
Created on 20 avr. 2021

@author: tleduc

Copyright 2020-2021 Thomas Leduc

This file is part of t4gpd.

t4gpd is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

t4gpd is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with t4gpd.  If not, see <https://www.gnu.org/licenses/>.
'''
from geopandas.geodataframe import GeoDataFrame
from t4gpd.commons.GeoProcess import GeoProcess
from t4gpd.commons.IllegalArgumentTypeException import IllegalArgumentTypeException
from t4gpd.commons.PointsDensifierLib import PointsDensifierLib


class STPointsDensifier(GeoProcess):
    '''
    classdocs
    '''

    def __init__(self, gdf, distance, pathidFieldname=None, adjustableDist=True,
                 removeDuplicate=True, distToTheSubstrate=None):
        '''
        Constructor
        '''
        if not isinstance(gdf, GeoDataFrame):
            raise IllegalArgumentTypeException(gdf, 'GeoDataFrame')
        if not ((pathidFieldname is None) or (pathidFieldname in gdf)):
            raise Exception(f'{pathidFieldname} is not a valid fieldname!')

        self.gdf = gdf
        self.pathidFieldname = pathidFieldname
        self.distance = distance
        self.adjustableDist = adjustableDist
        self.removeDuplicate = removeDuplicate
        self.distToTheSubstrate = distToTheSubstrate

    def run(self):
        rows = []

        if self.removeDuplicate:
            alreadyRegistered = set()
            for rowId, row in self.gdf.iterrows():
                geom = row.geometry
                if not self.pathidFieldname is None:
                    rowId = row[self.pathidFieldname]

                result = PointsDensifierLib.densifyByDistance(
                    geom, self.distance, rowId, contourid=0,
                    adjustableDist=self.adjustableDist,
                    distToTheSubstrate=self.distToTheSubstrate)

                for _row in result:
                    tmp = str(_row['geometry'])
                    if tmp not in alreadyRegistered:
                        alreadyRegistered.add(tmp)
                        rows.append(self.updateOrAppend(row, _row))

        else:
            for rowId, row in self.gdf.iterrows():
                geom = row.geometry
                if not self.pathidFieldname is None:
                    rowId = row[self.pathidFieldname]

                result = PointsDensifierLib.densifyByDistance(
                    geom, self.distance, rowId, contourid=0,
                    adjustableDist=self.adjustableDist,
                    distToTheSubstrate=self.distToTheSubstrate)

                for _row in result:
                    rows.append(self.updateOrAppend(row, _row))

        return GeoDataFrame(rows, crs=self.gdf.crs)



'''
Created on 28 sept. 2020

@author: tleduc

Copyright 2020 Thomas Leduc

This file is part of t4gpd.

t4gpd is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

t4gpd is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with t4gpd.  If not, see <https://www.gnu.org/licenses/>.
'''
from geopandas.geodataframe import GeoDataFrame
from shapely.geometry import Point
from t4gpd.commons.GeomLib import GeomLib
from t4gpd.commons.IllegalArgumentTypeException import IllegalArgumentTypeException
from t4gpd.commons.RayCasting2Lib import RayCasting2Lib
from t4gpd.commons.RayCastingLib import RayCastingLib
from t4gpd.commons.SVFLib import SVFLib
from t4gpd.morph.geoProcesses.AbstractGeoprocess import AbstractGeoprocess


class SkyViewFactor(AbstractGeoprocess):
    '''
    classdocs
    '''

    def __init__(self, buildingsGdf, nRays=64, maxRayLen=100.0, elevationFieldname='HAUTEUR',
                 method=2018, background=True):
        '''
        Constructor
        '''
        if not isinstance(buildingsGdf, GeoDataFrame):
            raise IllegalArgumentTypeException(buildingsGdf, 'GeoDataFrame')
        self.buildingsGdf = buildingsGdf
        self.spatialIndex = buildingsGdf.sindex

        self.shootingDirs = RayCastingLib.preparePanopticRays(nRays)
        self.maxRayLen = maxRayLen

        if elevationFieldname not in buildingsGdf:
            raise Exception('%s is not a relevant field name!' % elevationFieldname)
        self.elevationFieldname = elevationFieldname

        if (1981 == method):
            print(f'SVF calculation method: Oke (1981), nRays = {nRays}, maxRayLen = {maxRayLen:.1f}')
            self.method = SVFLib.svf1981
        else:
            print(f'SVF calculation method: Bernard et al. (2018), nRays = {nRays}, maxRayLen = {maxRayLen:.1f}')
            self.method = SVFLib.svf2018

        self.background = background

    def runWithArgs(self, row):
        viewPoint = row.geometry
        if not isinstance(viewPoint, Point):
            viewPoint = viewPoint.centroid

        if GeomLib.isAnIndoorPoint(viewPoint, self.buildingsGdf, self.spatialIndex):
            return { 'svf': 0.0 }

        '''
        _, _, hitDists, hitMasks, _ = RayCastingLib.multipleRayCast25D(
            self.buildingsGdf, self.spatialIndex, viewPoint, self.shootingDirs,
            self.maxRayLen, self.elevationFieldname, self.background)
        '''
        _, _, hitDists, hitMasks, _ = RayCasting2Lib.outdoorMultipleRayCast25D(
            self.buildingsGdf, self.spatialIndex, viewPoint, self.shootingDirs,
            self.maxRayLen, self.elevationFieldname, self.background)

        hitHeights = [0.0 if f is None else f[self.elevationFieldname] for f in hitMasks]

        return {
            # 'hit_dists': ArrayCoding.encode(hitDists),
            'svf': self.method(hitHeights, hitDists)
            }



'''
Created on 16 juin 2020

@author: tleduc

Copyright 2020 Thomas Leduc

This file is part of t4gpd.

t4gpd is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

t4gpd is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with t4gpd.  If not, see <https://www.gnu.org/licenses/>.
'''
from t4gpd.commons.GeoProcess import GeoProcess


class AbstractGeoprocess(GeoProcess):
    '''
    classdocs
    '''

    @staticmethod
    def runWithArgs(feature):
        raise NotImplementedError('runWithArgs(...) must be overridden!')



'''
Created on 19 juin 2020

@author: tleduc

Copyright 2020 Thomas Leduc

This file is part of t4gpd.

t4gpd is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

t4gpd is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with t4gpd.  If not, see <https://www.gnu.org/licenses/>.
'''
from shapely.ops import unary_union

from geopandas.geodataframe import GeoDataFrame

from t4gpd.commons.GeoProcess import GeoProcess
from t4gpd.commons.GeomLib import GeomLib
from t4gpd.commons.IllegalArgumentTypeException import IllegalArgumentTypeException


class STDilationErosion(GeoProcess):
    '''
    classdocs
    '''

    def __init__(self, inputGdf, buffDist):
        '''
        Constructor
        '''
        if not isinstance(inputGdf, GeoDataFrame):
            raise IllegalArgumentTypeException(inputGdf, 'GeoDataFrame')

        self.inputGdf = inputGdf
        self.buffDist = buffDist

    def run(self):
        sbuildings = list()
        for _, row in self.inputGdf.iterrows():
            envelop = GeomLib.removeHoles(row.geometry)
            if envelop is not None:
                sbuildings.append(envelop)
        unionOfSBuilding = unary_union(sbuildings)
        fullRegion = unionOfSBuilding.convex_hull

        ##### SQUARES IDENTIFICATION ####################
        uoSquares = fullRegion.difference(
            unionOfSBuilding.buffer(self.buffDist, -1)).buffer(self.buffDist, -1)
        
        rowsOfSquares = list()
        if GeomLib.isMultipart(uoSquares):
            for fid, g in enumerate(uoSquares.geoms):
                if GeomLib.isPolygonal(g):
                    rowsOfSquares.append({ 'FID': fid, 'geometry': g })
        else:
            rowsOfSquares.append({ 'FID': 0, 'geometry': uoSquares })
        squaresGdf = GeoDataFrame(rowsOfSquares, crs=self.inputGdf.crs)

        ##### STREETS IDENTIFICATION ####################
        uoStreets = fullRegion.difference(unionOfSBuilding).difference(uoSquares)

        rowsOfStreets = list()
        if (uoStreets is not None) and GeomLib.isMultipart(uoStreets):
            for fid, g in enumerate(uoStreets.geoms):
                if GeomLib.isPolygonal(g):
                    rowsOfStreets.append({ 'FID': fid, 'geometry': g })
        elif (uoStreets is not None):
            rowsOfStreets.append({ 'FID': 0, 'geometry': g })
        streetsGdf = GeoDataFrame(rowsOfStreets, crs=self.inputGdf.crs)

        return [ squaresGdf, streetsGdf ]



'''
Created on 28 avr. 2022

@author: tleduc

Copyright 2020-2022 Thomas Leduc

This file is part of t4gpd.

t4gpd is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

t4gpd is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with t4gpd.  If not, see <https://www.gnu.org/licenses/>.
'''
from datetime import datetime, timezone

from geopandas import GeoDataFrame
from numpy import cos, max, sin, tan
from shapely.geometry import Point
from t4gpd.commons.GeomLib import GeomLib
from t4gpd.commons.IllegalArgumentTypeException import IllegalArgumentTypeException
from t4gpd.commons.sun.SunLib import SunLib

from t4gpd.commons.RayCasting3Lib import RayCasting3Lib


class SunBeamLib(object):
    '''
    classdocs
    '''

    @staticmethod
    def isHitBySunBeam(position, dt, masks, maskElevationFieldname='HAUTEUR', model='pysolar'):
        if not isinstance(position, Point):
            raise IllegalArgumentTypeException(position, 'Point')
        if not isinstance(dt, datetime):
            raise IllegalArgumentTypeException(dt, 'datetime')
        if dt.tzinfo is None:
            dt = dt.replace(tzinfo=timezone.utc)
        if not isinstance(masks, GeoDataFrame):
            raise IllegalArgumentTypeException(masks, 'GeoDataFrame')
        masksSIdx = masks.sindex
        if maskElevationFieldname not in masks:
            raise Exception('%s is not a relevant field name!' % (maskElevationFieldname))

        ptA = position if GeomLib.is3D(position) else GeomLib.forceZCoordinateToZ0(position, z0=0.0)
        maxElevation = max(masks[maskElevationFieldname])
        sunModel = SunLib(masks, model)
        alti, azim = sunModel.getSolarAnglesInRadians(dt)
        R = (maxElevation - ptA.z) / tan(alti)
        ptB = Point([ptA.x + R * cos(azim), ptA.y + R * sin(azim), ptA.z + R * sin(alti)])

        # return RayCasting2Lib.areCovisible(ptA, ptB, masks, maskElevationFieldname, masksSIdx)
        return RayCasting3Lib.areCovisibleIn3D(ptA, ptB, masks, maskElevationFieldname, masksSIdx)



'''
Created on 21 juin 2021

@author: tleduc

Copyright 2020-2021 Thomas Leduc

This file is part of t4gpd.

t4gpd is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

t4gpd is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with t4gpd.  If not, see <https://www.gnu.org/licenses/>.
'''
from t4gpd.commons.GeomLib import GeomLib
from t4gpd.morph.geoProcesses.AbstractGeoprocess import AbstractGeoprocess


class GetInteriorPoint(AbstractGeoprocess):
    '''
    classdocs
    '''

    def runWithArgs(self, row):
        geom = row.geometry

        if 'Polygon' == geom.geom_type:
            return { 'geometry': GeomLib.getInteriorPoint(geom) }
        return {'geometry': None}



'''
Created on 1 avr. 2021

@author: tleduc

Copyright 2020-2021 Thomas Leduc

This file is part of t4gpd.

t4gpd is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

t4gpd is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with t4gpd.  If not, see <https://www.gnu.org/licenses/>.
'''
from t4gpd.morph.geoProcesses.AbstractGeoprocess import AbstractGeoprocess
from t4gpd.commons.GridFaceLib import GridFaceLib


class GridFace(AbstractGeoprocess):
    '''
    classdocs
    '''

    def __init__(self, dx, dy=None, intoPoint=True):
        '''
        Constructor
        '''
        self.dx = dx
        self.dy = dx if dy is None else dy
        self.intoPoint = intoPoint

    def runWithArgs(self, row):
        geom = row.geometry
        return GridFaceLib.grid(geom, self.dx, self.dy, self.intoPoint)



'''
Created on 11 mai 2021

@author: tleduc

Copyright 2020-2021 Thomas Leduc

This file is part of t4gpd.

t4gpd is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

t4gpd is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with t4gpd.  If not, see <https://www.gnu.org/licenses/>.
'''
from pandas.core.frame import DataFrame
from t4gpd.comfort.algo.TmrtGlobeTemperatureLib import TmrtGlobeTemperatureLib
from t4gpd.commons.IllegalArgumentTypeException import IllegalArgumentTypeException

from t4gpd.comfort.indices.AbstractThermalComfortIndice import AbstractThermalComfortIndice


class TmrtGlobeTemperature(AbstractThermalComfortIndice):
    '''
    classdocs
    '''

    def __init__(self, sensorsGdf, AirTC='Temperature', GlobeTC='GlobeTemperature',
                 WS_ms='WindSpeed'):
        '''
        Constructor

        AirTC: air temperature [C]
        Ws_ms: wind speed recorded at pedestrian level (at height 1.1 m) [m.s-1]
        '''
        if not isinstance(sensorsGdf, DataFrame):
            raise IllegalArgumentTypeException(sensorsGdf, 'DataFrame')

        for fieldname in (AirTC, GlobeTC, WS_ms):
            if fieldname not in sensorsGdf:
                raise Exception('%s is not a relevant field name!' % fieldname)

        self.AirTC = AirTC
        self.GlobeTC = GlobeTC
        self.WS_ms = WS_ms

    def runWithArgs(self, row):
        AirTC = row[self.AirTC]
        GlobeTC = row[self.GlobeTC]
        WS_ms = row[self.WS_ms]

        # calculation of the mean radiant temperature [C]
        # Mean radiant temperature after (Thorsson et al., 2007) based on observed globe temperature:
        T_mrt = TmrtGlobeTemperatureLib.assess_tmrt(AirTC, GlobeTC, WS_ms)

        return { 'T_mrt': T_mrt }



'''
Created on 3 feb. 2021

@author: tleduc

Copyright 2020 Thomas Leduc

This file is part of t4gpd.

t4gpd is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

t4gpd is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with t4gpd.  If not, see <https://www.gnu.org/licenses/>.
'''
from t4gpd.comfort.indices.AbstractThermalComfortIndice import AbstractThermalComfortIndice
from t4gpd.comfort.indices.ETU import ETU
from t4gpd.comfort.indices.PET import PET
from t4gpd.comfort.indices.PMV import PMV
from t4gpd.comfort.indices.PT import PT
from t4gpd.comfort.indices.SET import SET
from t4gpd.comfort.indices.SETmist import SETmist
from t4gpd.comfort.indices.TmrtOut import TmrtOut
from t4gpd.comfort.indices.UTCI import UTCI


class UniversalThermalIndices(AbstractThermalComfortIndice):
    '''
    classdocs
    '''

    def __init__(self, sensorsGdf, AirTC='AirTC_Avg', RH='RH_Avg', WS_ms='WS_ms_Avg', T_mrt='T_mrt',
                 N='N', Albedo='Albedo_1_Avg', SRUp='SR01Up_1_Avg', SRDn='SR01Dn_1_Avg',
                 IRUp='IR01UpCo_1_Avg', IRDn='IR01DnCo_1_Avg'):
        '''
        Constructor

        AirTC: air temperature [C]
        RH: relative humidity [%]
        Ws_ms: wind speed recorded at pedestrian level (at height 1.1 m) [m.s-1]
        T_mrt: Mean radiant temperature [C]
        N:
        Albedo:
        SRUp:
        SRDn:
        IRUp:
        IRDn:
        '''
        self.ops = [
            PET(sensorsGdf, AirTC, RH, WS_ms, T_mrt),
            PMV(sensorsGdf, AirTC, RH, WS_ms, T_mrt),
            PT(sensorsGdf, AirTC, RH, WS_ms, T_mrt),
            UTCI(sensorsGdf, AirTC, RH, WS_ms, T_mrt),
            SET(sensorsGdf, AirTC, RH, WS_ms, T_mrt),
            ETU(sensorsGdf, AirTC, RH, WS_ms, T_mrt, N, Albedo, SRUp, SRDn, IRUp, IRDn),
            TmrtOut(sensorsGdf, N, SRUp, SRDn, IRUp, IRDn),
            SETmist(sensorsGdf, AirTC, RH, WS_ms, T_mrt)
            ]

    def runWithArgs(self, row):
        result = dict()
        for op in self.ops:
            result.update(op.runWithArgs(row))
        return result



'''
Created on 19 juil. 2021

@author: tleduc

Copyright 2020-2021 Thomas Leduc

This file is part of t4gpd.

t4gpd is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

t4gpd is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with t4gpd.  If not, see <https://www.gnu.org/licenses/>.
'''
from glob import glob
from os import chdir, getcwd, makedirs
from os.path import basename, dirname
from tempfile import TemporaryDirectory
from zipfile import ZipFile

from geopandas import GeoDataFrame
from pandas import DataFrame
from t4gpd.commons.GeoProcess import GeoProcess
from t4gpd.commons.IllegalArgumentTypeException import IllegalArgumentTypeException


class ZipWriter(GeoProcess):
    '''
    classdocs
    '''
    DRIVER_EXT = {
        'ESRI Shapefile': 'shp',
        'GPKG': 'gpkg',
        'GeoJSON': 'geojson'
        }

    def __init__(self, mapOfDf, zipOutputFile, driver='ESRI Shapefile'):
        '''
        Constructor
        '''
        if not isinstance(mapOfDf, dict):
            raise IllegalArgumentTypeException(mapOfDf, '{ "layername": DataFrame, ... }')
        for layername, df in mapOfDf.items():
            if not isinstance(layername, str):
                raise IllegalArgumentTypeException(mapOfDf, '{ "layername": DataFrame, ... }')
            if not isinstance(df, DataFrame):
                raise IllegalArgumentTypeException(mapOfDf, '{ "layername": DataFrame, ... }')
        self.mapOfDf = mapOfDf

        self.zipOutputBasename = basename(zipOutputFile)
        self.zipOutputDirname = dirname(zipOutputFile)
        if (0 == len(self.zipOutputDirname)):
            self.zipOutputDirname = '.'

        if driver not in ('ESRI Shapefile', 'GPKG', 'GeoJSON'):
            raise IllegalArgumentTypeException(driver, '{"ESRI Shapefile", "GPKG", "GeoJSON"}')
        self.driver = driver

    def __zip(self, rootdir):
        with ZipFile(f'{self.zipOutputDirname}/{self.zipOutputBasename}.zip', 'w') as myzip:
            cwd = getcwd()
            chdir(rootdir)
            # writing each file one by one
            _ifiles = glob(f'{self.zipOutputBasename}/*.*', recursive=False)
            for _ifile in _ifiles:
                myzip.write(_ifile)
                # print(f'*** {_ifile}')  # DEBUG
            chdir(cwd)

    def run(self):
        extension = self.DRIVER_EXT[self.driver]
        with TemporaryDirectory() as tmpdir:
            makedirs(f'{tmpdir}/{self.zipOutputBasename}', exist_ok=True)
            for _layername, _df in self.mapOfDf.items():
                if not isinstance(_df, GeoDataFrame):
                    _df['geometry'] = None
                    _df = GeoDataFrame(_df)
                _ofile = f'{tmpdir}/{self.zipOutputBasename}/{_layername}.{extension}'
                if not _df.empty:
                    _df.to_file(_ofile, driver=self.driver, encoding='utf-8')
                # print(f'=== {_ofile}')  # DEBUG

            self.__zip(tmpdir)

        print(f'{self.zipOutputDirname}/{self.zipOutputBasename}.zip has been written!')
        return None



'''
Created on 31 mai 2022

@author: tleduc

Copyright 2020-2022 Thomas Leduc

This file is part of t4gpd.

t4gpd is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

t4gpd is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with t4gpd.  If not, see <https://www.gnu.org/licenses/>.
'''
import cv2
from geopandas import GeoDataFrame, overlay
from numpy import zeros
from shapely.geometry import box
from t4gpd.commons.GeoProcess import GeoProcess
from t4gpd.commons.IllegalArgumentTypeException import IllegalArgumentTypeException


class STToRaster(GeoProcess):
    '''
    classdocs
    '''

    def __init__(self, gdf, nrows, ncols, bbox=None, threshold=0.5, outputFile=None):
        '''
        Constructor
        '''
        if not isinstance(gdf, GeoDataFrame):
            raise IllegalArgumentTypeException(gdf, 'GeoDataFrame')
        self.gdf = gdf

        self.nrows, self.ncols = nrows, ncols
        self.minx, self.miny, maxx, self.maxy = gdf.total_bounds if bbox is None else bbox
        self.dx = (maxx - self.minx) / ncols
        self.dy = (self.maxy - self.miny) / nrows
        self.threshold = threshold
        self.outputFile = outputFile

    def __grid(self):
        result = []
        x0 = self.minx
        for c in range(self.ncols):
            x0pp, y0 = x0 + self.dx, self.maxy
            for r in range(self.nrows):
                y0pp = y0 - self.dy
                result.append({
                    'gid': r * self.ncols + c,
                    'row': r,
                    'col': c,
                    'geometry': box(x0, y0, x0pp, y0pp)
                })
                y0 = y0pp
            x0 = x0pp
        return GeoDataFrame(result, crs=self.gdf.crs)

    def __fromGridToArray(self, grid):
        img = zeros([self.nrows, self.ncols], dtype=int)
        for _, rowItem in grid.iterrows():
            if (self.threshold <= rowItem.cover):
                img[rowItem.row, rowItem.col] = 1
        return img

    def run(self):
        grid = self.__grid()

        grid = overlay(grid, self.gdf[['geometry']], how='intersection')
        grid = grid.dissolve(by='gid', as_index=False)
        cellArea = self.dx * self.dy
        grid['cover'] = grid.geometry.apply(lambda g: g.area / cellArea)

        img = self.__fromGridToArray(grid)

        if not self.outputFile is None:
            cv2.imwrite(self.outputFile, 255 * img)

        return img



'''
Created on 11 mai 2021

@author: tleduc

Copyright 2020-2021 Thomas Leduc

This file is part of t4gpd.

t4gpd is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

t4gpd is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with t4gpd.  If not, see <https://www.gnu.org/licenses/>.
'''
from pandas.core.frame import DataFrame
from t4gpd.comfort.algo.ConstantsLib import ConstantsLib
from t4gpd.commons.IllegalArgumentTypeException import IllegalArgumentTypeException

from t4gpd.comfort.indices.AbstractThermalComfortIndice import AbstractThermalComfortIndice


class TmrtRadiometer(AbstractThermalComfortIndice):
    '''
    classdocs
    '''

    def __init__(self, sensorsGdf, SRUp='SR01Up_1_Avg', SRDn='SR01Dn_1_Avg',
                 SRright='SR01Up_2_Avg', SRleft='SR01Dn_2_Avg',
                 SRfront='SR01Up_3_Avg', SRback='SR01Dn_3_Avg',
                 IRUp='IR01UpCo_1_Avg', IRDn='IR01DnCo_1_Avg',
                 IRright='IR01UpCo_2_Avg', IRleft='IR01DnCo_2_Avg',
                 IRfront='IR01UpCo_3_Avg', IRback='IR01DnCo_3_Avg'):
        '''
        Constructor
        '''
        if not isinstance(sensorsGdf, DataFrame):
            raise IllegalArgumentTypeException(sensorsGdf, 'DataFrame')

        for fieldname in (SRUp, SRDn, SRright, SRleft, SRfront, SRback,
                          IRUp, IRDn, IRright, IRleft, IRfront, IRback):
            if fieldname not in sensorsGdf:
                raise Exception('%s is not a relevant field name!' % fieldname)

        self.SRUp, self.SRDn = SRUp, SRDn
        self.SRright, self.SRleft = SRright, SRleft
        self.SRfront, self.SRback = SRfront, SRback

        self.IRUp, self.IRDn = IRUp, IRDn
        self.IRright, self.IRleft = IRright, IRleft
        self.IRfront, self.IRback = IRfront, IRback

    def runWithArgs(self, row):
        SRUp, SRDn = row[self.SRUp], row[self.SRDn]
        SRright, SRleft = row[self.SRright], row[self.SRleft]
        SRfront, SRback = row[self.SRfront], row[self.SRback]

        IRUp, IRDn = row[self.IRUp], row[self.IRDn] 
        IRright, IRleft = row[self.IRright], row[self.IRleft]
        IRfront, IRback = row[self.IRfront], row[self.IRback]

        # calculation of the mean flux density of radiation
        S_rad = (ConstantsLib.alpha_k * (
            ConstantsLib.F_vert * (SRUp + SRDn) + 
            ConstantsLib.F_lat * (SRright + SRleft + SRfront + SRback)) + 
            ConstantsLib.alpha_l * (
                ConstantsLib.F_vert * (IRUp + IRDn) + 
                ConstantsLib.F_lat * (IRright + IRleft + IRfront + IRback)))

        # 14.04.2021
        # Mean flux densities splitted in 6 directions according Kantor et al. (2016)
        _SRUp = (ConstantsLib.alpha_k * ConstantsLib.F_vert * SRUp)
        _SRDn = (ConstantsLib.alpha_k * ConstantsLib.F_vert * SRDn)
        _IRUp = (ConstantsLib.alpha_l * ConstantsLib.F_vert * IRUp)
        _IRDn = (ConstantsLib.alpha_l * ConstantsLib.F_vert * IRDn)
        _SRright = (ConstantsLib.alpha_k * ConstantsLib.F_lat * SRright)
        _SRleft = (ConstantsLib.alpha_k * ConstantsLib.F_lat * SRleft)
        _SRfront = (ConstantsLib.alpha_k * ConstantsLib.F_lat * SRfront)
        _SRback = (ConstantsLib.alpha_k * ConstantsLib.F_lat * SRback)
        _IRright = (ConstantsLib.alpha_l * ConstantsLib.F_lat * IRright)
        _IRleft = (ConstantsLib.alpha_l * ConstantsLib.F_lat * IRleft)
        _IRfront = (ConstantsLib.alpha_l * ConstantsLib.F_lat * IRfront)
        _IRback = (ConstantsLib.alpha_l * ConstantsLib.F_lat * IRback)
        _SRlat = (ConstantsLib.alpha_k * ConstantsLib.F_lat * (SRright + SRleft + SRfront + SRback))
        _IRlat = (ConstantsLib.alpha_l * ConstantsLib.F_lat * (IRright + IRleft + IRfront + IRback))

        # calculation of the mean radiant temperature [C]
        T_mrt = ((S_rad / (ConstantsLib.epsi_p * ConstantsLib.sigma_B)) ** (0.25) - ConstantsLib.T_kel)

        # T_mrt splitted in up, down and lateral IR and SR according to Middel and Krayenhoff (2019) and Kantor et al. (2016)
        _ratio = T_mrt / S_rad

        T_mrt_SRUp = _ratio * _SRUp
        T_mrt_SRDn = _ratio * _SRDn
        T_mrt_IRUp = _ratio * _IRUp
        T_mrt_IRDn = _ratio * _IRDn
        T_mrt_SRright = _ratio * _SRright
        T_mrt_SRleft = _ratio * _SRleft
        T_mrt_SRfront = _ratio * _SRfront
        T_mrt_SRback = _ratio * _SRback
        T_mrt_IRright = _ratio * _IRright
        T_mrt_IRleft = _ratio * _IRleft
        T_mrt_IRfront = _ratio * _IRfront
        T_mrt_IRback = _ratio * _IRback
        T_mrt_SRlat = _ratio * _SRlat
        T_mrt_IRlat = _ratio * _IRlat

        return { 
            'T_mrt': T_mrt,
            #
            'T_mrt_SRUp': T_mrt_SRUp,
            'T_mrt_SRDn': T_mrt_SRDn,
            'T_mrt_IRUp': T_mrt_IRUp,
            'T_mrt_IRDn': T_mrt_IRDn,
            #
            'T_mrt_SRright': T_mrt_SRright,
            'T_mrt_SRleft': T_mrt_SRleft,
            'T_mrt_SRfront': T_mrt_SRfront,
            'T_mrt_SRback': T_mrt_SRback,
            #
            'T_mrt_IRright': T_mrt_IRright,
            'T_mrt_IRleft': T_mrt_IRleft,
            'T_mrt_IRfront': T_mrt_IRfront,
            'T_mrt_IRback': T_mrt_IRback,
            #
            'T_mrt_SRlat': T_mrt_SRlat,
            'T_mrt_IRlat': T_mrt_IRlat
            }



'''
Created on 12 mai 2021

@author: tleduc

Copyright 2020-2021 Thomas Leduc

This file is part of t4gpd.

t4gpd is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

t4gpd is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with t4gpd.  If not, see <https://www.gnu.org/licenses/>.
'''
from pythermalcomfort.models import set_tmp
try:
    from pythermalcomfort.psychrometrics import v_relative
except ImportError:
    from pythermalcomfort.utilities import v_relative

from t4gpd.comfort.algo.ConstantsLib import ConstantsLib


class SETLib(object):
    '''
    classdocs
    '''

    @staticmethod
    def assess_set(AirTC, RH, WS_ms, T_mrt):
        # https://pythermalcomfort.readthedocs.io

        # Estimates the relative air speed which combines the average air speed of the
        # space plus the relative air speed caused by the body movement.
        vr = v_relative(WS_ms, ConstantsLib.M_met)

        # Calculates the Standard Effective Temperature (SET). The SET is the temperature 
        # of an imaginary environment at 50% (rh), <0.1 m/s (20 fpm) average air speed (v),
        # and tr = tdb , in which the total heat loss from the skin of an imaginary 
        # occupant with an activity level of 1.0 met and a clothing level of 0.6 clo is the
        # same as that from a person in the actual environment with actual clothing and 
        # activity level.
        tdb, tr, v, rh, met, clo = (AirTC, T_mrt, vr, RH, ConstantsLib.M_met, ConstantsLib.Clo)
        SET = set_tmp(tdb, tr, v, rh, met, clo, wme=0,
                      body_surface_area=1.8258, p_atm=101325,
                      units='SI')
        return SET



'''
Created on 18 dec. 2020

@author: tleduc

Copyright 2020 Thomas Leduc

This file is part of t4gpd.

t4gpd is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

t4gpd is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with t4gpd.  If not, see <https://www.gnu.org/licenses/>.
'''
from t4gpd.morph.geoProcesses.AbstractGeoprocess import AbstractGeoprocess


class ConvexityIndices(AbstractGeoprocess):
    '''
    classdocs
    '''

    @staticmethod
    def runWithArgs(row):
        geom = row.geometry
        geomArea = geom.area
        geomPerim = geom.length

        chull = geom.convex_hull
        chullArea = chull.area
        chullPerim = chull.length

        connectedComponents = chull.difference(geom).geoms

        nConnectedComponents = len(connectedComponents)
        areaConvexityDefect = geomArea / chullArea if (0.0 < chullArea) else None
        perimConvexityDefect = chullPerim / geomPerim if (0.0 < geomPerim) else None

        bigConcavities = sum([g.area ** 2 for g in connectedComponents]) / nConnectedComponents
        smallConcavities = sum([g.area ** (-2) 
                                for g in connectedComponents
                                if (0.0 < g.area)]) / nConnectedComponents

        return {
            'n_con_comp': nConnectedComponents,
            'a_conv_def': areaConvexityDefect,
            'p_conv_def': perimConvexityDefect,
            'big_concav': bigConcavities,
            'small_conc': smallConcavities
            }        



'''
Created on 2 feb. 2021

@author: tleduc

Copyright 2020 Thomas Leduc

This file is part of t4gpd.

t4gpd is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

t4gpd is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with t4gpd.  If not, see <https://www.gnu.org/licenses/>.
'''
from t4gpd.comfort.indices.ASV import ASV
from t4gpd.comfort.indices.AbstractThermalComfortIndice import AbstractThermalComfortIndice


class EmpiricalThermalIndices(AbstractThermalComfortIndice):
    '''
    classdocs
    '''

    def __init__(self, sensorsGdf, AirTC='AirTC_Avg', RH='RH_Avg', WS_ms='WS_ms_Avg',
                 SRUp='SR01Up_1_Avg'):
        '''
        Constructor

        AirTC: air temperature [C]
        RH: relative humidity [%]
        Ws_ms: wind speed [m.s-1]
        SRUp: incoming shortwave radiation [W.m-2]
        '''
        self.ops = [
            ASV(sensorsGdf, AirTC, RH, WS_ms, SRUp)
            ]

    def runWithArgs(self, row):
        result = dict()
        for op in self.ops:
            result.update(op.runWithArgs(row))
        return result



'''
Created on 29 sept. 2020

@author: tleduc

Copyright 2020 Thomas Leduc

This file is part of t4gpd.

t4gpd is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

t4gpd is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with t4gpd.  If not, see <https://www.gnu.org/licenses/>.
'''
from geopandas.geodataframe import GeoDataFrame
from numpy import mean
from t4gpd.commons.GeomLib import GeomLib
from t4gpd.commons.IllegalArgumentTypeException import IllegalArgumentTypeException
from t4gpd.commons.RayCastingLib import RayCastingLib
from t4gpd.morph.geoProcesses.AbstractGeoprocess import AbstractGeoprocess


class AspectRatio(AbstractGeoprocess):
    '''
    classdocs
    '''
    DIRECTIONAL = 'Directional'
    PANOPTIC = 'Panoptic'
    MIXED = 'Mixed'

    def __init__(self, sensorsGdf, buildingsGdf, nRays=64, maxRayLen=100.0, modality='Panoptic',
                 elevationFieldname='HAUTEUR', canyonFieldName=None):
        '''
        Constructor
        '''
        if modality in (self.DIRECTIONAL, self.PANOPTIC, self.MIXED):
            if not isinstance(sensorsGdf, GeoDataFrame):
                raise IllegalArgumentTypeException(sensorsGdf, 'GeoDataFrame')

            if not isinstance(buildingsGdf, GeoDataFrame):
                raise IllegalArgumentTypeException(buildingsGdf, 'GeoDataFrame')
            self.buildingsGdf = buildingsGdf
            self.spatialIndex = buildingsGdf.sindex
    
            if (2 > nRays):
                raise IllegalArgumentTypeException(nRays, 'nRays must be greater than 2!')
            self.shootingDirs = RayCastingLib.preparePanopticRays(nRays)
            self.maxRayLen = maxRayLen

            self.modality = modality
            if (self.MIXED == modality) and (
                (canyonFieldName is None) or (canyonFieldName not in sensorsGdf)):
                raise Exception('If you chose the "mixed" modality, you must set "canyonFieldName"!')

            if elevationFieldname not in buildingsGdf:
                raise Exception('%s is not a relevant field name!' % elevationFieldname)
            self.elevationFieldname = elevationFieldname

            self.canyonFieldName = canyonFieldName
        else:
            raise Exception('Illegal argument: modality must be chosen in {%s, %s, %s}!' % 
                            (self.DIRECTIONAL, self.PANOPTIC, self.MIXED))

    def runWithArgs(self, row):
        viewPoint = row.geometry.centroid

        if GeomLib.isAnIndoorPoint(viewPoint, self.buildingsGdf, self.spatialIndex):
            return { 'svf': 0.0 }

        _shootingDirs = self.shootingDirs

        if (self.DIRECTIONAL == self.modality):
            _shootingDirs = RayCastingLib.prepareOrientedRays(
                self.buildingsGdf, self.spatialIndex, viewPoint)
        elif ((self.MIXED == self.modality) and
              ((1 == row[self.canyonFieldName]) or row[self.canyonFieldName])):
            _shootingDirs = RayCastingLib.prepareOrientedRays(
                self.buildingsGdf, self.spatialIndex, viewPoint)

        _, _, hitDists, hitMasks, _ = RayCastingLib.multipleRayCast25D(
            self.buildingsGdf, self.spatialIndex, viewPoint, _shootingDirs,
            self.maxRayLen, self.elevationFieldname, background=False)

        hitHeights = [0.0 if f is None else f[self.elevationFieldname] for f in hitMasks]

        # When the length is infinite, fix it by default at the distance to the artificial horizon
        hitDists = [x if (float('inf') != x) else self.maxRayLen for x in hitDists]

        return {
            # 'geometry': rays,
            # 'hit_dists': ArrayCoding.encode(hitDists),
            'h_over_w': float(mean(hitHeights) / (2 * mean(hitDists)))
            }



'''
Created on 3 mars 2021

@author: tleduc

Copyright 2020 Thomas Leduc

This file is part of t4gpd.

t4gpd is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

t4gpd is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with t4gpd.  If not, see <https://www.gnu.org/licenses/>.
'''
from geopandas.geodataframe import GeoDataFrame
from t4gpd.commons.IllegalArgumentTypeException import IllegalArgumentTypeException
from t4gpd.commons.sun.SunLib import SunLib
from t4gpd.sun.geoProcesses.AbstractSunshineDuration import AbstractSunshineDuration


class SunshineDuration(AbstractSunshineDuration):
    '''
    classdocs
    '''

    def __init__(self, masksGdf, maskElevationFieldname, datetimes, model='pysolar'):
        '''
        Constructor
        '''
        if not isinstance(masksGdf, GeoDataFrame):
            raise IllegalArgumentTypeException(masksGdf, 'GeoDataFrame')
        self.masksGdf = masksGdf
        self.masksSIdx = masksGdf.sindex

        if maskElevationFieldname not in masksGdf:
            raise Exception('%s is not a relevant field name!' % (maskElevationFieldname))
        self.maskElevationFieldname = maskElevationFieldname
        maxElevation = max(self.masksGdf[self.maskElevationFieldname])  #===========

        self.sunModel = SunLib(masksGdf, model)
        self.sunPositions = self._getAllSunPositions(datetimes, maxElevation)
        self.nSunPositions = len(self.sunPositions)

    def runWithArgs(self, row):
        viewpoint = row.geometry.centroid

        nbHits = 0
        for (sunAlti, radDir, rayLen) in self.sunPositions:
            if self._beingInTheSun(viewpoint, radDir, rayLen, sunAlti):
                nbHits += 1

        return {
            'sun_hits': nbHits,
            'sun_ratio': float(nbHits) / self.nSunPositions
            }



'''
Created on 17 juin 2020

@author: tleduc

Copyright 2020 Thomas Leduc

This file is part of t4gpd.

t4gpd is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

t4gpd is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with t4gpd.  If not, see <https://www.gnu.org/licenses/>.
'''
from shapely.geometry import LineString, Point
from t4gpd.commons.graph.AbstractUrbanGraphLib import AbstractUrbanGraphLib


class UrbanGraphLib(AbstractUrbanGraphLib):
    '''
    classdocs
    '''

    def __init__(self, roads):
        '''
        Constructor
        '''
        self.roads, self.graph, self.ciVertices, self.icVertices, self.edges = AbstractUrbanGraphLib.initializeGraph(roads)
        self.spatialIdx = self.roads.sindex

    def getUniqueRoadsSectionsNodes(self):
        result = list()
        for k, v in self.edges.items():
            nbConnectedVertices = len(v)
            if 2 != nbConnectedVertices:
                result.append({'geometry': Point(self.icVertices[k]), 'gid': k,
                               'nb_connections': nbConnectedVertices })
        return result

    def __getNextVertex(self, prev, curr):
        succ = list(self.edges[curr])
        if 2 == len(succ):
            if (prev == succ[0]) and not (prev == succ[1]):
                return succ[1]
            elif (prev == succ[1]) and not (prev == succ[0]):
                return succ[0]
        return None

    def __burnEdge(self, idx1, idx2, burnedEdges):
        minIdx = min(idx1, idx2)
        maxidx = max(idx1, idx2)
        if minIdx in burnedEdges:
            burnedEdges[minIdx].add(maxidx)
        else:
            burnedEdges[minIdx] = set({ maxidx })

    def __buildRoadSection(self, vertexIndex, nextVertexIndex, burnedEdges):
        result = []

        prevIdx = vertexIndex
        result.append(self.icVertices[prevIdx])
        currIdx = nextVertexIndex
        result.append(self.icVertices[currIdx])
        self.__burnEdge(prevIdx, currIdx, burnedEdges)

        nextIdx = self.__getNextVertex(prevIdx, currIdx)
        while nextIdx is not None:
            result.append(self.icVertices[nextIdx])
            prevIdx = currIdx
            currIdx = nextIdx
            self.__burnEdge(prevIdx, currIdx, burnedEdges)

            if (nextIdx == vertexIndex):
                # current road section is a loop, stop the process
                nextIdx = None
            else:
                nextIdx = self.__getNextVertex(prevIdx, currIdx)            

        return LineString(result)

    def getUniqueRoadsSections(self):
        result = []
        burnedEdges = dict()
        for vertexIndex in list(self.edges.keys()):
            nbConnectedVertices = len(self.edges[vertexIndex])
            if 2 != nbConnectedVertices:
                # current vertex is a 'boundary' of a 'road section'
                for nextVertexIndex in self.edges[vertexIndex]:
                    minIdx = min(vertexIndex, nextVertexIndex)
                    maxIdx = max(vertexIndex, nextVertexIndex)
                    if not (minIdx in burnedEdges and (maxIdx in burnedEdges[minIdx])):
                        geom = self.__buildRoadSection(vertexIndex, nextVertexIndex, burnedEdges)
                        result.append({ 'geometry': geom, 'gid': len(result),
                                       'distance': geom.length })
        return result

    def getUniqueRoadsSectionsWithoutCulDeSac(self):
        result = []

        rs = self.getUniqueRoadsSections()
        for row in rs:
            _coords = list(row['geometry'].coords)
            startHash = UrbanGraphLib.hashCoord(_coords[0])
            stopHash = UrbanGraphLib.hashCoord(_coords[-1])
            if ((1 < len(self.edges[self.ciVertices[startHash]])) and
                (1 < len(self.edges[self.ciVertices[stopHash]]))):
                result.append(row)
        return result



'''
Created on 22 août 2022

@author: tleduc

Copyright 2020-2022 Thomas Leduc

This file is part of t4gpd.

t4gpd is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

t4gpd is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with t4gpd.  If not, see <https://www.gnu.org/licenses/>.
'''
from geopandas import GeoDataFrame
from numpy import asarray, dot, where
from shapely.geometry import MultiLineString, Point
from t4gpd.commons.IllegalArgumentTypeException import IllegalArgumentTypeException
from t4gpd.morph.geoProcesses.AbstractGeoprocess import AbstractGeoprocess
from t4gpd.pyvista.ToUnstructuredGrid import ToUnstructuredGrid
from t4gpd.pyvista.commons.RayCasting3DLib import RayCasting3DLib
from t4gpd.pyvista.commons.RayCastingIn3DLib import RayCastingIn3DLib


class SVF3D(AbstractGeoprocess):
    '''
    classdocs
    '''

    def __init__(self, listOfMasks, nrays=1000, method='MonteCarlo'):
        '''
        Constructor
        '''
        assert isinstance(listOfMasks, (list, tuple)), 'listOfMasks must be a list of GeoDataFrames'
        assert all(isinstance(gdf, GeoDataFrame) for gdf in listOfMasks), 'listOfMasks must be a list of GeoDataFrames'
        masks = ToUnstructuredGrid(listOfMasks).run()
        self.maxRayLen = masks.length

        self.obbTree = RayCastingIn3DLib.prepareVtkOBBTree(masks)

        self.nrays = nrays
        self.main_direction = asarray([0, 0, 1])
        self.method = method

        self.shootingDirs = None
        if not ('MonteCarlo' == self.method):
            rays = RayCasting3DLib.preparePanopticRays(2 * nrays, method)
            self.shootingDirs = self.maxRayLen * self.__selectUpwardFacingRays(rays)

    def __selectUpwardFacingRays(self, rays):
        dotProducts = dot(rays, self.main_direction)
        indices = where(dotProducts >= 0)[0]
        return rays[indices,:]

    def runWithArgs(self, row):
        geom = row.geometry
        if not isinstance(geom, Point):
            raise IllegalArgumentTypeException(geom, 'GeoDataFrame of Point')

        if ('MonteCarlo' == self.method):
            rays = RayCasting3DLib.prepareOrientedRandomRays(
                self.nrays, self.main_direction, openness=90, method='MonteCarlo')
            shootingDirs = self.maxRayLen * rays
        else:
            shootingDirs = self.shootingDirs

        srcPt = asarray(geom.coords[0])
        dstPts = srcPt + shootingDirs
        srcPts = srcPt.reshape(1, -1).repeat(len(dstPts), axis=0)

        rays, _, hitDists, _ = RayCastingIn3DLib.mraycastObbTree(self.obbTree, srcPts, dstPts)

        return {
            # 'geometry': MultiLineString(rays),
            'svf': len([d for d in hitDists if (d == RayCastingIn3DLib.INFINITY)]) / len(hitDists)
            }



'''
Created on 12 mai 2021

@author: tleduc

Copyright 2020-2021 Thomas Leduc

This file is part of t4gpd.

t4gpd is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

t4gpd is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with t4gpd.  If not, see <https://www.gnu.org/licenses/>.
'''
from numpy import exp

from t4gpd.commons.AngleLib import AngleLib


class ConstantsLib(object):
    '''
    classdocs
    '''
    # ah [-] solar absorptivity of the clothed body after Watanabe et al. (2013) casual clothing:
    # summer = 0.66
    # autumn = 0.69
    # winter = 0.77
    ah = 0.66

    alpha = 0.2  # alpha 0.2 for sedentary, 0.1 for vasodilation and 0.33 for vasoconstriction
    alpha_eff = 0.72  # effective radiation surface of a subject in upright standing position
    alpha_k = 0.7  # absorption coefficient for shortwave radiation (visible = SR)
    alpha_l = 0.97  # absorption coefficient for longwave radiation (infrared = IR)

    body_surface_area = 1.8258  # body surface area, default value 1.8258 [m2] 

    Clo = 0.5  # clothing level

    # parameters for (Kestrel) Tmrt calculation based on observed globe temperature 
    D_glo = 25  # Diameter of globe thermometer (mm)
    epsi_p = 0.97  # body emissivity
    epsi_glo = 0.95  # Emissivite du globe for a black globe (Khuen et al., 1970)

    # fa_eff --> effective area factor between 0.1 and 1 (Oh et al., 2020) 
    # --> constant set at 0.5
    fa_eff = 0.5

    # clothing surface area factor fcl depending on clothing level 
    fcl = (1.0 + 0.2 * Clo) if (0.5 >= Clo) else (1.05 + 0.1 * Clo)  # clothing surface area factor

    # Clo clothing level in standard environment is 0 so that:
    # Fcleo effective clothing thermal efficiency [-] in standard environment is 1.0
    Fcleo = 1.0
    # fclo clothing area factor in the standard environment, dimensionless selon Watanabe et al. (2014)
    fclo = 0.31

    # for radiation calculation in ETU
    # fp [-] projected area factor
    fp = alpha_eff

    F_lat = 0.22  # angular factor for lateral direction (N-S et E-W)
    F_vert = 0.06  # angular factor for vertical direction (up-down)

    # hco convective heat transfer coefficient [W.m-2.K-1] in standard environment 3.0 W.m-2.K-1
    hco = 3.0
    # heo evaporative heat transfer coefficient [W.m-2.kPa-1] in standard environment 3.0 LR W/(m2 kPa)
    heo = 3.0

    hfg = 2430  # hfg heat of vaporization of water [kJ.kg-1] at 30 C

    # hro ou hrc radiant heat transfer coefficient [W.m-2.K-1] in standard environment 4.5 W.m-2.K-1
    hro = 4.5

    # h_sun solar altitude [deg]
    h_sun = 66
    # function to convert angles from deg to rad
    h_sun_rad = AngleLib.toRadians(h_sun) 

    icl = Clo  # intrinsic permeation efficiency ratio
    Icl = Clo * 0.155  # clothing insulation, in square meters Kelvin per watt (m2 K/W)

    Iclo = fclo * 0.155

    M = 93  # metabolic rate M, in Watt per square meter (W/m2)
    M_met = M / 58.2  # metabolic rate M_met, in met

    patm = 101325  # atmospheric pressure, default value 101325 [Pa]

    # phi_skyh [-] angle factor between the human body and sky surface
    phi_skyh = F_vert 
    # phi_groundh [-] angle factor between the human body and ground surface
    phi_groundh = F_vert

    p_mist = 0.25  # default value in Oh et al (2020)

    sigma_B = 5.67e-8  # Wm^(-2) K^(-4) Stefan-Boltzman constant

    tcr = 37.0  # body core temperature [C]
    tcrn = 36.8  # neutral core temperature [C]
    tskn = 33.7  # neutral skin temperature [C]

    T_kel = 273.15  # temperature in Kelvin for temperature conversion

    # T_v hypothetical air temperature [C] in standard environment: the air temperature of an 
    # isothermal environment at predicted mean vote (PMV)=0 
    T_v = 27.1

    W = 0  # effective mechanical power, in Watt per square meter (W/m2)
    W_met = W / 58.2  # effective mechanical power, in met

    tsk = 35.7 - 0.0275 * (M - W)  # skin external temperature tsk depending on M and W

    # Fpclo permeation efficiency factor in the standard environment, dimensionless
    Fpclo = 1 / (1 + 0.143 * hco * Iclo)

    # vapor pressure at skin temperature tsk
    # saturated air pressure at skin
    P_ssk = exp(18.956 - (4030.183 / (tsk + 235)))  # [mb] 10mb =1kPa = 1000Pa = 1000N.m-2
    P_ssk = P_ssk / 10  # [kPa]



'''
Created on 15 dec. 2020

@author: tleduc

Copyright 2020 Thomas Leduc

This file is part of t4gpd.

t4gpd is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

t4gpd is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with t4gpd.  If not, see <https://www.gnu.org/licenses/>.
'''
from shapely.geometry import LineString, Polygon

from t4gpd.commons.GeomLib import GeomLib
from t4gpd.commons.IllegalArgumentTypeException import IllegalArgumentTypeException


class CaliperLib(object):
    '''
    classdocs
    '''

    def __init__(self, criteria=None):
        '''
        Constructor
        '''
        if criteria in (None, 'min', 'max'):
            self.maximize = True if criteria in [None, 'max'] else False
        else:
            raise IllegalArgumentTypeException(criteria, '{None, "min", "max"}')

    @staticmethod
    def __getDistanceToAntipodalVertex(pairOfVertices, vertices):
        listOfQuintuplets = CaliperLib.__getListOfQuintuplets(pairOfVertices, vertices)
        return max([dist for _, _, _, _, dist in listOfQuintuplets])

    @staticmethod
    def __getListOfQuintuplets(pairOfVertices, vertices):
        result = list()
        for vertex in vertices:
            # projP, distFromAToProjP, distFromPToLine = GeomUtilities.projectOntoStraightLine(vertex, pairOfVertices)
            result.append([vertex, pairOfVertices] + GeomLib.projectOntoStraightLine(vertex, pairOfVertices))
        return result

    @staticmethod
    def __getListOfListOfQuintuplets(geom):
        vertices, nVertices = CaliperLib.__preprocess(geom)
        result = list()
        for i in range(nVertices):
            nextI = (i + 1) % nVertices
            v1, v2 = vertices[i], vertices[nextI]
            result.append(CaliperLib.__getListOfQuintuplets([v1, v2], vertices))
        return result

    @staticmethod
    def __getListOfMaxOfQuintuplets(geom):
        listOfListOfQuintuplets = CaliperLib.__getListOfListOfQuintuplets(geom)
        result = list()
        for listOfQuintuplets in listOfListOfQuintuplets:
            maxDist = -1.0
            for quintuplet in listOfQuintuplets:
                if (maxDist < quintuplet[4]):
                    maxDist, currResult = quintuplet[4], quintuplet
            result.append(currResult)
        return result

    @staticmethod
    def __getEnclosingRectangle(pairOfVertices, vertices):
        listOfQuintuplets = CaliperLib.__getListOfQuintuplets(pairOfVertices, vertices)

        a, b = pairOfVertices
        u = GeomLib.unitVector(a, b)
        v = [-u[1], u[0]]

        minDu = min([distFromAToProjP for _, _, _, distFromAToProjP, _ in listOfQuintuplets])
        maxDu = max([distFromAToProjP for _, _, _, distFromAToProjP, _ in listOfQuintuplets])
        maxDv = max([distFromPToLine for _, _, _, _, distFromPToLine in listOfQuintuplets])

        p1 = [ a[0] + u[0] * minDu, a[1] + u[1] * minDu ]
        p2 = [ a[0] + u[0] * maxDu, a[1] + u[1] * maxDu ]
        p3 = [ p2[0] + v[0] * maxDv, p2[1] + v[1] * maxDv ]
        p4 = [ p1[0] + v[0] * maxDv, p1[1] + v[1] * maxDv ]

        return [Polygon([p1, p2, p3, p4, p1]), max(maxDu - minDu, maxDv), min(maxDu - minDu, maxDv)]

    @staticmethod
    def __getAntipodalVertex(pairOfVertices, vertices):
        listOfQuintuplets = CaliperLib.__getListOfQuintuplets(pairOfVertices, vertices)

        maxDist, result = -1.0, None
        for p, _, _, _, distFromPToLine in listOfQuintuplets:
            if (maxDist < distFromPToLine):
                maxDist, result = distFromPToLine, p
        return maxDist, result

    @staticmethod
    def __getDistances(geom):
        listOfListOfQuintuplets = CaliperLib.__getListOfListOfQuintuplets(geom)
        distances = [
            max([dist for _, _, _, _, dist in listOfQuintuplets])
            for listOfQuintuplets in listOfListOfQuintuplets
            ]
        return distances

    @staticmethod
    def __preprocess(geom):
        convexHull = geom.convex_hull
        if convexHull is None:
            return None, 0
        # vertices = list(convexHull.exterior.coords)[:-1]  # Clockwise order
        vertices = list(reversed(list(convexHull.exterior.coords)[:-1]))  # Counter-clockwise order
        return vertices, len(vertices)

    def diameter(self, geom):
        listOfMaxOfQuintuplets = CaliperLib.__getListOfMaxOfQuintuplets(geom)

        pairOfVertices, oppVertex, projOppVertex = None, None, None
        if self.maximize:
            maxDist = -1.0
            for p, pairOfV, projP, _, dist in listOfMaxOfQuintuplets:
                if (maxDist < dist):
                    maxDist, pairOfVertices, oppVertex, projOppVertex = dist, pairOfV, p, projP
        else:
            minDist = float('inf')
            for p, pairOfV, projP, _, dist in listOfMaxOfQuintuplets:
                if (minDist > dist):
                    minDist, pairOfVertices, oppVertex, projOppVertex = dist, pairOfV, p, projP
        # return [pairOfVertices, [oppVertex, projOppVertex]]
        return [oppVertex, projOppVertex]

    def length(self, geom):
        distances = CaliperLib.__getDistances(geom)
        if self.maximize:
            return max(distances)
        return min(distances)

    def orientation(self, geom):
        pairOfVertices = self.diameter(geom)
        if pairOfVertices is None:
            return None
        segment = LineString(pairOfVertices)
        return float(GeomLib.getLineStringOrientation(segment))

    @staticmethod
    def stretching(geom):
        distances = CaliperLib.__getDistances(geom)
        minDist, maxDist = min(distances), max(distances)
        if (0.0 >= maxDist):
            return None
        return float(minDist / maxDist)

    @staticmethod
    def getBoundingRectangles(geom):
        vertices, nVertices = CaliperLib.__preprocess(geom)
        result = list()
        for i in range(nVertices):
            nextI = (i + 1) % nVertices
            v1, v2 = vertices[i], vertices[nextI]
            result.append(CaliperLib.__getEnclosingRectangle([v1, v2], vertices))
        return result

    @staticmethod
    def mabr(geom):
        rectangles = CaliperLib.getBoundingRectangles(geom)
        minArea, result = float('inf'), None
        for rect, len1, len2 in rectangles:
            currArea = len1 * len2
            if currArea < minArea:
                minArea = currArea
                result = [rect, len1, len2]
        return result

    @staticmethod
    def mpbr(geom):
        rectangles = CaliperLib.getBoundingRectangles(geom)
        minHalfPerim, result = float('inf'), None
        for rect, len1, len2 in rectangles:
            currHalfPerim = len1 + len2
            if currHalfPerim < minHalfPerim:
                minHalfPerim = currHalfPerim
                result = [rect, len1, len2]
        return result



'''
Created on 12 mai 2021

@author: tleduc

Copyright 2020-2021 Thomas Leduc

This file is part of t4gpd.

t4gpd is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

t4gpd is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with t4gpd.  If not, see <https://www.gnu.org/licenses/>.
'''
from pandas.core.frame import DataFrame
from t4gpd.comfort.algo.PMVLib import PMVLib
from t4gpd.commons.IllegalArgumentTypeException import IllegalArgumentTypeException

from t4gpd.comfort.indices.AbstractThermalComfortIndice import AbstractThermalComfortIndice


class PMV(AbstractThermalComfortIndice):
    '''
    classdocs
    '''

    def __init__(self, sensorsGdf, AirTC='AirTC_Avg', RH='RH_Avg', WS_ms='WS_ms_Avg',
                 T_mrt='T_mrt'):
        '''
        Constructor

        AirTC: air temperature [C]
        RH: relative humidity [%]
        Ws_ms: wind speed recorded at pedestrian level (at height 1.1 m) [m.s-1]
        T_mrt: Mean radiant temperature [C]
        
        PMV: Predicted Mean Vote
        '''
        if not isinstance(sensorsGdf, DataFrame):
            raise IllegalArgumentTypeException(sensorsGdf, 'DataFrame')

        for fieldname in (AirTC, RH, WS_ms, T_mrt):
            if fieldname not in sensorsGdf:
                raise Exception('%s is not a relevant field name!' % fieldname)

        self.AirTC = AirTC
        self.RH = RH
        self.WS_ms = WS_ms
        self.T_mrt = T_mrt

    def runWithArgs(self, row):
        AirTC = row[self.AirTC]
        RH = row[self.RH]
        WS_ms = row[self.WS_ms]
        T_mrt = row[self.T_mrt]

        # PMV: Predicted Mean Vote
        PMV = PMVLib.assess_pmv(AirTC, RH, WS_ms, T_mrt)

        return { 'PMV': PMV }



'''
Created on 14 dec. 2020

@author: tleduc

Copyright 2020 Thomas Leduc

This file is part of t4gpd.

t4gpd is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

t4gpd is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with t4gpd.  If not, see <https://www.gnu.org/licenses/>.
'''
from geopandas.geodataframe import GeoDataFrame
from shapely.ops import unary_union
from t4gpd.commons.GeoProcess import GeoProcess
from t4gpd.commons.GeomLib import GeomLib
from t4gpd.commons.IllegalArgumentTypeException import IllegalArgumentTypeException


class STExtractOpenSpaces(GeoProcess):
    '''
    classdocs
    '''
    EPSILON = 1e-3

    def __init__(self, envelopeGdf, buildings, removeCourtyards=False):
        '''
        Constructor
        '''
        if not isinstance(envelopeGdf, GeoDataFrame):
            raise IllegalArgumentTypeException(envelopeGdf, 'GeoDataFrame')
        self.envelopeGdf = envelopeGdf

        if not isinstance(buildings, GeoDataFrame):
            raise IllegalArgumentTypeException(buildings, 'GeoDataFrame')
        self.buildings = buildings

        self.removeCourtyards = removeCourtyards

    def run(self):
        # Use a buffer to avoid slivers
        xunion = lambda gdf: unary_union([g.buffer(self.EPSILON) for g in gdf.geometry]).buffer(self.EPSILON)

        envelope = xunion(self.envelopeGdf)
        buildings = xunion(self.buildings)
        if self.removeCourtyards:
            buildings = GeomLib.removeHoles(buildings)

        result = envelope.difference(buildings)
        # Use a buffer to avoid slivers
        result = result.buffer(self.EPSILON)

        return GeoDataFrame([{'geometry':result}], crs=self.buildings.crs)



'''
Created on 18 juil. 2022

@author: tleduc

Copyright 2020-2022 Thomas Leduc

This file is part of t4gpd.

t4gpd is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

t4gpd is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with t4gpd.  If not, see <https://www.gnu.org/licenses/>.
'''
from geopandas import GeoDataFrame
from t4gpd.commons.GeomLib import GeomLib
from t4gpd.commons.GeomLib3D import GeomLib3D
from t4gpd.morph.geoProcesses.AbstractGeoprocess import AbstractGeoprocess
from t4gpd.pyvista.commons.RayCastingIn3DLib import RayCastingIn3DLib
from t4gpd.pyvista.ToUnstructuredGrid import ToUnstructuredGrid
from shapely.geometry import LinearRing, LineString, Polygon
from t4gpd.commons.IllegalArgumentTypeException import IllegalArgumentTypeException


class AutomaticFaceOrientation(AbstractGeoprocess):
    '''
    classdocs
    '''

    def __init__(self, gdf, dist=1e-3):
        '''
        Constructor
        '''
        assert isinstance(gdf, GeoDataFrame), 'gdf must be a GeoDataFrame'
        self.obbTree = RayCastingIn3DLib.prepareVtkOBBTree(gdf)
        self.maxHeight = ToUnstructuredGrid([gdf]).run().length
        self.dist = dist

    @staticmethod
    def __reverseOrientation(obj):
        if isinstance(obj, LineString):
            return LinearRing(reversed(obj.coords))

        elif isinstance(obj, Polygon):
            extRing = AutomaticFaceOrientation.__reverseOrientation(obj.exterior)
            intRings = [AutomaticFaceOrientation.__reverseOrientation(hole) for hole in obj.interiors]
            return Polygon(extRing, intRings)

        raise IllegalArgumentTypeException(obj, 'Shapely LineString or Polygon')

    def runWithArgs(self, row):
        geom = row.geometry
        assert GeomLib.isPolygonal(geom), 'The GeoDataFrame must be composed of (Multi)Polygon'
        c = GeomLib3D.centroid(geom).coords[0]
        n = GeomLib3D.getFaceNormalVector(geom)

        # The objective here is to move the Ray Casting point away from the face
        srcPt = [ c[i] + self.dist * n[i] for i in range(3) ]
        dstPt = [ srcPt[0], srcPt[1], srcPt[2] + self.maxHeight ]
        if RayCastingIn3DLib.areCovisibleObbTree(self.obbTree, srcPt, dstPt):
            # FACE IS CORRECTLY ORIENTED
            return { 'reverse': 0 }

        srcPt = [ c[i] - self.dist * n[i] for i in range(3) ]
        dstPt = [ srcPt[0], srcPt[1], srcPt[2] + self.maxHeight ]
        if RayCastingIn3DLib.areCovisibleObbTree(self.obbTree, srcPt, dstPt):
            # FACE IS NOT CORRECTLY ORIENTED
            geom = AutomaticFaceOrientation.__reverseOrientation(geom)
            return { 'geometry': geom, 'reverse': 1 }

        # INDECISION
        return { 'reverse': 2 }



'''
Created on 27 juil. 2022

@author: tleduc

Copyright 2020-2022 Thomas Leduc

This file is part of t4gpd.

t4gpd is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

t4gpd is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with t4gpd.  If not, see <https://www.gnu.org/licenses/>.
'''
from geopandas.geodataframe import GeoDataFrame
from t4gpd.commons.GeoProcess import GeoProcess
from t4gpd.commons.IllegalArgumentTypeException import IllegalArgumentTypeException
from t4gpd.pyqgis.AddMemoryLayer import AddMemoryLayer


class Emphasizer(GeoProcess):
    '''
    classdocs
    '''

    def __init__(self, layer, fieldname, fieldvalues, bufferRadius):
        '''
        Constructor
        '''
        if not isinstance(layer, GeoDataFrame):
            raise IllegalArgumentTypeException(layer, 'GeoDataFrame')

        if fieldname not in layer:
            raise Exception(f'{fieldname} is not a relevant field name!')

        self.layer = layer[ layer[fieldname].isin(fieldvalues) ].copy(deep=True)
        self.layer.geometry = self.layer.geometry.apply(
            lambda g: g.centroid.buffer(bufferRadius)) 

    def run(self):
        return AddMemoryLayer(self.layer, 'highlighted layer').run()



'''
Created on 3 mars 2021

@author: tleduc

Copyright 2020 Thomas Leduc

This file is part of t4gpd.

t4gpd is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

t4gpd is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with t4gpd.  If not, see <https://www.gnu.org/licenses/>.
'''
from geopandas.geodataframe import GeoDataFrame
from numpy import arctan
from shapely.geometry import LineString, Point
from t4gpd.commons.GeomLib import GeomLib
from t4gpd.commons.IllegalArgumentTypeException import IllegalArgumentTypeException
from t4gpd.commons.sun.SunLib import SunLib
from t4gpd.sun.geoProcesses.AbstractSunshineDuration import AbstractSunshineDuration


class SunshineDurationOnTopOfRoof(AbstractSunshineDuration):
    '''
    classdocs
    '''

    def __init__(self, masksGdf, maskElevationFieldname, datetimes, model='pysolar'):
        '''
        Constructor
        '''
        if not isinstance(masksGdf, GeoDataFrame):
            raise IllegalArgumentTypeException(masksGdf, 'GeoDataFrame')
        self.masksGdf = masksGdf
        self.masksSIdx = masksGdf.sindex

        if maskElevationFieldname not in masksGdf:
            raise Exception('%s is not a relevant field name!' % (maskElevationFieldname))
        self.maskElevationFieldname = maskElevationFieldname
        maxElevation = max(self.masksGdf[self.maskElevationFieldname])  #===========

        self.sunModel = SunLib(masksGdf, model)
        self.sunPositions = self._getAllSunPositions(datetimes, maxElevation)
        self.nSunPositions = len(self.sunPositions)
 
    @staticmethod
    def __beingInTheSunOnTheRooftop(masksGdf, masksSIdx, maskElevationFieldname, viewpoint,
                                    h0, radDir, rayLen, sunAlti):
        # To avoid: "Inconsistent coordinate dimensionality"
        viewpoint = Point((viewpoint.x, viewpoint.y))
        remotePoint = Point((viewpoint.x + rayLen * radDir[0], viewpoint.y + rayLen * radDir[1]))
        sunRay = LineString([viewpoint, remotePoint])

        masksIds = list(masksSIdx.intersection(sunRay.bounds))
        for maskId in masksIds:
            mask = masksGdf.loc[maskId]
            maskGeom = mask.geometry
            maskElevation = mask[maskElevationFieldname]

            if sunRay.crosses(maskGeom):
                tmpGeom = maskGeom.intersection(sunRay)
                gc = tmpGeom.geoms if GeomLib.isMultipart(tmpGeom) else [tmpGeom]

                for geom in gc:
                    for vertex in geom.coords:
                        hitPoint = Point(vertex)
                        _dist = viewpoint.distance(hitPoint)
                        if (h0 < maskElevation):
                            _angle = arctan((maskElevation - h0) / _dist)
                            if (_angle >= sunAlti):
                                return False

        return True

    def runWithArgs(self, row):
        # viewpoint = row.geometry.centroid
        viewpoint = row.geometry.representative_point()

        enclosingFeatures = GeomLib.getEnclosingFeatures(self.masksGdf, self.masksSIdx, viewpoint)

        if (0 == len(enclosingFeatures)):
            # OUTDOOR/STREET-LEVEL VIEWPOINT
            nbHits = 0
            for (sunAlti, radDir, rayLen) in self.sunPositions:
                if self._beingInTheSun(viewpoint, radDir, rayLen, sunAlti):
                    nbHits += 1

        else:
            # INDOOR/ROOFTOP VIEWPOINT

            # REMOVE enclosingFeatures FROM THE GeoDataFrame & SpatialIndex OF SURROUNDING MASKS
            enclosingFeaturesIndex = [f.name for f in enclosingFeatures]
            _buildings = self.masksGdf[~self.masksGdf.index.isin(enclosingFeaturesIndex)]
            _buildings.reset_index(drop=True, inplace=True)  # HANDLE POTENTIAL KeyError
            _sindex = _buildings.sindex

            # ASSESS SUN BEAM HITS
            h0 = min([f[self.maskElevationFieldname] for f in enclosingFeatures])

            nbHits = 0
            for (sunAlti, radDir, rayLen) in self.sunPositions:
                if SunshineDurationOnTopOfRoof.__beingInTheSunOnTheRooftop(
                    _buildings, _sindex, self.maskElevationFieldname, viewpoint,
                    h0, radDir, rayLen, sunAlti):
                    nbHits += 1

        return {
            'sun_hits': nbHits,
            'sun_ratio': float(nbHits) / self.nSunPositions
            }



'''
Created on 12 mai 2021

@author: tleduc

Copyright 2020-2021 Thomas Leduc

This file is part of t4gpd.

t4gpd is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

t4gpd is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with t4gpd.  If not, see <https://www.gnu.org/licenses/>.
'''
from pythermalcomfort.models import pmv_ppd
try:
    from pythermalcomfort.psychrometrics import v_relative
except ImportError:
    from pythermalcomfort.utilities import v_relative
from t4gpd.comfort.algo.ConstantsLib import ConstantsLib


class PMVLib(object):
    '''
    classdocs
    '''

    @staticmethod
    def assess_pmv(AirTC, RH, WS_ms, T_mrt):
        # https://pythermalcomfort.readthedocs.io

        # Estimates the relative air speed which combines the average air speed of the
        # space plus the relative air speed caused by the body movement.
        vr = v_relative(WS_ms, ConstantsLib.M_met)

        # The PMV is an index that predicts the mean value of the thermal sensation votes (self-
        # reported perceptions) of a large group of people on a sensation scale expressed from -3
        # to +3 corresponding to the categories "cold", "cool", "slightly cool", "neutral", 
        # "slightly warm", "warm", and "hot".
        tdb, tr, vr, rh, met, clo = (AirTC, T_mrt, vr, RH, ConstantsLib.M_met, ConstantsLib.Clo)
        _result = pmv_ppd(tdb, tr, vr, rh, met, clo, wme=0, standard='ISO', units='SI')

        return _result['pmv']



'''
Created on 2 avr. 2021

@author: tleduc

Copyright 2020-2021 Thomas Leduc

This file is part of t4gpd.

t4gpd is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

t4gpd is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with t4gpd.  If not, see <https://www.gnu.org/licenses/>.
'''
from geopandas.geodataframe import GeoDataFrame
from shapely.geometry import LineString, MultiPolygon
from t4gpd.commons.AngleLib import AngleLib
from t4gpd.commons.GeoProcess import GeoProcess
from t4gpd.commons.GeomLib import GeomLib
from t4gpd.commons.IllegalArgumentTypeException import IllegalArgumentTypeException


class STFacadesAnalysis(GeoProcess):
    '''
    classdocs
    '''

    def __init__(self, buildings, gidFieldname='ID', elevationFieldname='HAUTEUR', exteriorOnly=True):
        '''
        Constructor
        '''
        if not isinstance(buildings, GeoDataFrame):
            raise IllegalArgumentTypeException(buildings, 'GeoDataFrame')
        self.buildings = buildings

        if not ((gidFieldname is None) or (gidFieldname in buildings)):
            raise Exception('%s is not a relevant field name!' % (gidFieldname))
        self.gidFieldname = gidFieldname

        if not ((elevationFieldname is None) or (elevationFieldname in buildings)):
            raise Exception('%s is not a relevant field name!' % (elevationFieldname))
        self.elevationFieldname = elevationFieldname

        self.exteriorOnly = exteriorOnly

    def __analyzeFacades(self, geom, gid, elev):
        rows = []

        if isinstance(geom, MultiPolygon):
            for _geom in geom.geoms:
                rows += self.__analyzeFacades(_geom, gid, elev)
        else:
            if self.exteriorOnly:
                geom = geom.exterior
            for _currEdge in GeomLib.toListOfBipointsAsLineStrings(geom):
                _midPoint = GeomLib.getMidPoint(*_currEdge.coords)
                _normal = GeomLib.unitVector(*_currEdge.coords)
                _normal = [_normal[1], -_normal[0]]
                _azim = AngleLib.toDegrees(AngleLib.normAzimuth(_normal))
                _normal = [_midPoint[0] + _normal[0], _midPoint[1] + _normal[1]]
                _normal = LineString([_midPoint, _normal])
                _length = _currEdge.length
                rows.append({ 
                    'buildingID': None if (gid is None) else gid,
                    'length': _length,
                    'height': None if (elev is None) else elev,
                    'surface': None if (elev is None) else _length * elev,
                    'azimuth': _azim,
                    'normal': _normal.wkt,
                    'geometry': _currEdge
                    })

        return rows

    def run(self):
        rows = []
        for _, row in self.buildings.iterrows():
            if not GeomLib.isPolygonal(row.geometry):
                continue            
            gid = None if (self.gidFieldname is None) else row[self.gidFieldname] 
            elev = None if (self.elevationFieldname is None) else row[self.elevationFieldname] 
            geom = GeomLib.normalizeRingOrientation(row.geometry)
            rows += self.__analyzeFacades(geom, gid, elev)

        result = GeoDataFrame(rows, crs=self.buildings.crs)
        result['gid'] = range(len(result))
        return result



'''
Created on 18 dec. 2020

@author: tleduc

Copyright 2020 Thomas Leduc

This file is part of t4gpd.

t4gpd is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

t4gpd is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with t4gpd.  If not, see <https://www.gnu.org/licenses/>.
'''
from t4gpd.morph.geoProcesses.AbstractGeoprocess import AbstractGeoprocess
from t4gpd.commons.CaliperLib import CaliperLib


class RectangularityIndices(AbstractGeoprocess):
    '''
    classdocs
    '''

    @staticmethod
    def runWithArgs(row):
        geom = row.geometry
        geomArea = geom.area
        geomPerim = geom.length

        # mabr = geom.minimum_rotated_rectangle
        mabr, len1, len2 = CaliperLib().mabr(geom)
        mabrArea = mabr.area
        mabrPerim = mabr.length

        stretching = min(len1, len2) / max(len1, len2) if (0.0 < max(len1, len2)) else None
        areaRectDefect = geomArea / mabrArea if (0.0 < mabrArea) else None
        perimRectDefect = mabrPerim / geomPerim if (0.0 < geomPerim) else None

        return {
            'stretching': stretching,
            'a_rect_def': areaRectDefect,
            'p_rect_def': perimRectDefect
            }



'''
Created on 25 aug 2020

@author: tleduc

Copyright 2020 Thomas Leduc

This file is part of t4gpd.

t4gpd is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

t4gpd is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with t4gpd.  If not, see <https://www.gnu.org/licenses/>.
'''
from numpy import cos, pi, sin


class Perez(object):
    '''
    classdocs
    voir ~/prj/solenetb/src/trunk/solene.core.orig/luminance_ciel.c
    irradiance du rayonnement solaire extra-atmospherique sur une surface perpendiculaire au rayon
    '''
    SOLAR_CONSTANT = 1367.0
    '''
    http://atmospheres.gsfc.nasa.gov/climate/index.php?section=136
    The total solar irradiance (TSI), improperly called "solar constant" until a few years ago,
    has been found to change about 0.1% in an 11-year solar sunspot activity. The current most
    accurate TSI values from the Total Irradiance Monitor (TIM) on NASA's Solar Radiation and
    Climate Experiment ( SORCE ) is 1360.8 +/- 0.5 W/m2 during the 2008 solar minimum as compared
    to previous estimates of 1365.4 +/- 1.3 W/m2 established in the 1990s.
    '''
    TOTAL_SOLAR_IRRADIANCE = 1360.8

    @staticmethod
    def __dogniauxOpticalAirMass(solarAltitudeAngle):
        # ~ https://en.wikipedia.org/wiki/Air_mass_%28astronomy%29
        # ~ d'apres (Miguet, 2000; p. 164)
        y = (3.885 + solarAltitudeAngle) ** (-1.253)
        z = sin(solarAltitudeAngle)
        return (1.0 / (z + 0.15 * y))

    @staticmethod
    def __eccentricity(dayInYear):
        # ~ d'apres (Miguet, 2000; p. 162)
        k = (2 * pi * (dayInYear - 1)) / 365.0
        return (1.00011 + 0.034221 * cos(k) + 0.00128 * sin(k) + 0.000719 * cos(2 * k) + 0.000077 * sin(2 * k))

    @staticmethod
    def diffuseSolarIrradiance(solarAltitudeAngle, dayInYear, delta):
        # ~ delta: sky's brightness (Perez et al., 1993)
        assert (0 <= solarAltitudeAngle <= (pi / 2)), 'solarAltitudeAngle in radians!'

        airMass = Perez.__dogniauxOpticalAirMass(solarAltitudeAngle)
        eccentricity = Perez.__eccentricity(dayInYear)
        return ((delta * Perez.SOLAR_CONSTANT * eccentricity) / airMass)

    @staticmethod
    def directSolarIrradiance(solarAltitudeAngle, dayInYear, delta, epsilon):
        # ~ delta: sky's brightness (Perez et al., 1993)
        # ~ epsilon: sky's clearness (Perez et al., 1993)
        assert (0 <= solarAltitudeAngle <= (pi / 2)), 'solarAltitudeAngle in radians!'

        de = Perez.diffuseSolarIrradiance(solarAltitudeAngle, dayInYear, delta)
        z = pi / 2.0 - solarAltitudeAngle
        return de * (epsilon - 1.0) * (1.0 + 1.041 * z ** 3)



'''
Created on 8 oct. 2020

@author: tleduc

Copyright 2020 Thomas Leduc

This file is part of t4gpd.

t4gpd is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

t4gpd is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with t4gpd.  If not, see <https://www.gnu.org/licenses/>.
'''
from builtins import isinstance
from datetime import datetime
from warnings import warn

from geopandas.geodataframe import GeoDataFrame
from shapely.geometry import LineString, Polygon
from t4gpd.commons.GeoProcess import GeoProcess
from t4gpd.commons.GeomLib import GeomLib
from t4gpd.commons.IllegalArgumentTypeException import IllegalArgumentTypeException


class ObjWriter(GeoProcess):
    '''
    classdocs
    '''

    def __init__(self, inputGdf, outputFile):
        '''
        Constructor
        '''
        if not isinstance(inputGdf, GeoDataFrame):
            raise IllegalArgumentTypeException(inputGdf, 'GeoDataFrame')
        self.inputGdf = inputGdf

        self.outputFile = outputFile

    def __dumpVertex(self, coords):
        z = 0.0
        if 2 == len(coords):
            x, y = coords
        elif 3 == len(coords):
            x, y, z = coords
        return 'v %.5f %.5f %.5f\n' % (x, y, z)
        # return 'v %g %g %g\n' % (x, y, z)

    def __dumpListOfVertices(self, myDictOfVertices, vertices):
        return ' '.join([str(myDictOfVertices[vtx]) for vtx in vertices])

    def run(self):
        myDictOfVertices, myListOfLines, myListOfFaces = dict(), [], []

        for _, row in self.inputGdf.iterrows():
            geom = row.geometry

            points = GeomLib.getListOfShapelyPoints(geom)
            for _p in points:
                p = _p.coords[0]
                if p not in myDictOfVertices:
                    myDictOfVertices[p] = 1 + len(myDictOfVertices)

            if isinstance(geom, Polygon):
                myListOfFaces.append(list(geom.exterior.coords))
            elif isinstance(geom, LineString):
                myListOfLines.append(geom.coords)
            else:
                warn('ObjWriter deals only with LineString and Polygon!')

        with open(self.outputFile, 'w') as out:
            out.write('# t4gpd.io.ObjWriter - dev. in 2020, by T. Leduc, AAU-CRENAU\n')
            out.write('# OBJ generated on %s\n\n' % (datetime.now().strftime("%d.%m.%Y at %H:%M:%S")))
            out.write('# ===== %d vertices =====\n\n' % (len(myDictOfVertices)))

            for p, _ in sorted(myDictOfVertices.items(), key=lambda kv:kv[1]):
                out.write(self.__dumpVertex(p))

            if 0 < len(myListOfFaces):
                out.write('\n# ===== %d face(s) =====\n' % (len(myListOfFaces)))
                for i, f in enumerate(myListOfFaces, start=1):
                    out.write('\ng %d\n' % (i))
                    out.write('f %s\n' % (self.__dumpListOfVertices(myDictOfVertices, f)))

            if 0 < len(myListOfLines):
                out.write('\n# ===== %d polyline(s) =====\n\n' % (len(myListOfLines)))
                for l in myListOfLines:
                    out.write('l %s\n' % (self.__dumpListOfVertices(myDictOfVertices, l)))



'''
Created on 9 sept. 2021

@author: tleduc

Copyright 2020-2021 Thomas Leduc

This file is part of t4gpd.

t4gpd is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

t4gpd is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with t4gpd.  If not, see <https://www.gnu.org/licenses/>.
'''
from geopandas.geodataframe import GeoDataFrame
from t4gpd.commons.GeoProcess import GeoProcess
from t4gpd.commons.IllegalArgumentTypeException import IllegalArgumentTypeException

from qgis.PyQt.Qt import QVariant
from qgis.core import QgsCoordinateReferenceSystem, QgsFeature, QgsField, QgsGeometry, QgsProject, QgsVectorLayer


class AddMemoryLayer(GeoProcess):
    '''
    classdocs
    '''

    def __init__(self, gdf, layername):
        '''
        Constructor
        '''
        if not isinstance(gdf, GeoDataFrame):
            raise IllegalArgumentTypeException(gdf, 'GeoDataFrame')
        self.gdf = gdf

        if not isinstance(layername, str):
            raise IllegalArgumentTypeException(layername, 'str')
        self.layername = layername

    @staticmethod
    def __fromGeoDataFrameToQgsVectorLayer(gdf, layername):
        geomTypes = gdf.geom_type.unique()
        if (1 == len(geomTypes)):
            geomType = geomTypes[0]
        elif (2 == len(geomTypes)):
            geomType = geomTypes[0] if (str(geomTypes[0]).startswith('Multi')) else geomTypes[1]

        qgsVectorLayer = QgsVectorLayer(geomType, layername, 'memory')
        provider = qgsVectorLayer.dataProvider()
        qgsVectorLayer.startEditing()
        fieldnames, qgsFields = [], []
        for _fieldname, _fieldtype in gdf.dtypes.items():
            if not ('geometry' == str(_fieldtype)):
                _fieldtype = QVariant.String
                if (str(_fieldtype).startswith('int')):
                    _fieldtype = QVariant.Int
                elif (str(_fieldtype).startswith('float')):
                    _fieldtype = QVariant.Double
                elif (str(_fieldtype).startswith('datetime')):
                    _fieldtype = QVariant.DateTime
                fieldnames.append(_fieldname)
                qgsFields.append(QgsField(_fieldname, QVariant.String))
        provider.addAttributes(qgsFields)

        qgsFeatures = []
        for _, row in gdf.iterrows():
            qgsFeature = QgsFeature()
            qgsFeature.setGeometry(QgsGeometry.fromWkt(row.geometry.wkt))
            qgsFeature.setAttributes([row[_f] for _f in fieldnames])
            qgsFeatures.append(qgsFeature)
        provider.addFeatures(qgsFeatures)

        # qgsVectorLayer.setCrs(QgsCoordinateReferenceSystem(gdf.crs.to_epsg()))
        qgsVectorLayer.setCrs(QgsCoordinateReferenceSystem.fromEpsgId(gdf.crs.to_epsg()))
        qgsVectorLayer.commitChanges()
        qgsVectorLayer.updateExtents()
        return qgsVectorLayer

    def run(self):
        qgsVectorLayer = self.__fromGeoDataFrameToQgsVectorLayer(self.gdf, self.layername)
        QgsProject.instance().addMapLayer(qgsVectorLayer)
        return qgsVectorLayer



'''
Created on 2 oct. 2020

@author: tleduc

Copyright 2020 Thomas Leduc

This file is part of t4gpd.

t4gpd is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

t4gpd is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with t4gpd.  If not, see <https://www.gnu.org/licenses/>.
'''
from geopandas import GeoDataFrame
from numpy import inf
from pandas import concat
from shapely.geometry import LineString
from shapely.ops import unary_union
from t4gpd.commons.GeoProcess import GeoProcess
from t4gpd.commons.GeomLib import GeomLib
from t4gpd.commons.IllegalArgumentTypeException import IllegalArgumentTypeException
from t4gpd.morph.STPointsDensifier import STPointsDensifier
from t4gpd.morph.STVoronoiPartition import STVoronoiPartition


class STCoolscapesTessellation(GeoProcess):
    '''
    classdocs
    '''

    def __init__(self, linesGdf, sampleDist, buffDist, magnitude=5):
        '''
        Constructor
        '''
        if not isinstance(linesGdf, GeoDataFrame):
            raise IllegalArgumentTypeException(linesGdf, 'GeoDataFrame')

        self.linesGdf = linesGdf
        self.pathGeom = unary_union(self.linesGdf.geometry)
        if not isinstance(self.pathGeom, LineString):
            raise IllegalArgumentTypeException(linesGdf, 'GeoDataFrame of continuous LineStrings')

        self.sampleDist = sampleDist
        self.buffDist = buffDist
        self.magnitude = magnitude

    def __getSensors(self):
        _sensors = STPointsDensifier(
            self.linesGdf, self.sampleDist, pathidFieldname=None, adjustableDist=True,
            removeDuplicate=True).run()

        _bigBuffZoneGeom = self.pathGeom.buffer(self.magnitude * self.buffDist)
        _bigBuffZone = GeoDataFrame([{'geometry': _bigBuffZoneGeom}], crs=self.linesGdf.crs)

        _otherSensors = STPointsDensifier(
            _bigBuffZone, self.magnitude * self.sampleDist, pathidFieldname=None,
            adjustableDist=True, removeDuplicate=True).run()

        return concat([_sensors, _otherSensors])

    def __orderTessellationCells(self, tessellation):
        mydict = dict()
        for _, rowCell in tessellation.iterrows():
            cellGid = rowCell['gid']
            cellCurvAbsc = self.pathGeom.project(rowCell.geometry.centroid)
            mydict[cellGid] = float(cellCurvAbsc)

        orddict = sorted(mydict.items(), key=lambda kv: kv[1])
        correspTable = { cellGid:i for i, (cellGid, _) in enumerate(orddict)}
        tessellation['gid'] = tessellation['gid'].apply(lambda cellGid: correspTable[cellGid])
        tessellation.sort_values(by='gid', inplace=True)
        return tessellation

    def run(self):
        _tessellation = STVoronoiPartition(self.__getSensors()).run()
        _buffZoneGeom = self.pathGeom.buffer(self.buffDist)

        _tessellation.geometry = _tessellation.geometry.apply(lambda g: g.intersection(_buffZoneGeom))

        # FILTER EMPTY GEOMETRIES
        _tessellation = _tessellation[~_tessellation.is_empty]
        # ORDER TESSELLATION CELLS
        _tessellation = self.__orderTessellationCells(_tessellation)

        # ADD START/STOP CURVILINEAR ABSCISSA TO EACH TESSELLATION CELL
        startCurvAbsc, stopCurvAbsc = [], []

        nCells = len(_tessellation)
        for _, row in _tessellation.iterrows():
            _gid = row['gid']
            _tmpGeom = row.geometry.boundary.intersection(self.pathGeom)
            _tmpGeom = GeomLib.getListOfShapelyPoints(_tmpGeom)
            _tmpCurvAbsc = sorted([self.pathGeom.project(p) for p in _tmpGeom])

            if (0 == _gid):
                startCurvAbsc.append(-inf)
                stopCurvAbsc.append(_tmpCurvAbsc[-1])
            elif (nCells - 1 == _gid):
                startCurvAbsc.append(stopCurvAbsc[-1])
                stopCurvAbsc.append(inf)
            else:
                if (2 != len(_tmpCurvAbsc)):
                    raise Exception('STCoolscapesTessellation:: Unreachable instruction!')
                startCurvAbsc.append(_tmpCurvAbsc[0])
                stopCurvAbsc.append(_tmpCurvAbsc[1])

        _tessellation['curv_abs_0'] = startCurvAbsc
        _tessellation['curv_abs_1'] = stopCurvAbsc

        return _tessellation



'''
Created on 3 juil. 2020

@author: tleduc

Copyright 2020 Thomas Leduc

This file is part of t4gpd.

t4gpd is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

t4gpd is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with t4gpd.  If not, see <https://www.gnu.org/licenses/>.
'''

from geopandas.geodataframe import GeoDataFrame
from shapely.geometry import Point
from t4gpd.commons.CSVLib import CSVLib
from t4gpd.commons.GeoProcess import GeoProcess


class CSVReader(GeoProcess):
    '''
    classdocs
    '''

    def __init__(self, inputFile, xFieldName='longitude', yFieldName='latitude',
                 fieldSep=',', srcEpsgCode='EPSG:4326', dstEpsgCode=None, decimalSep='.'):
        '''
        Constructor
        '''
        self.inputFile = inputFile
        self.xFieldName = xFieldName
        self.yFieldName = yFieldName
        self.fieldSep = fieldSep
        self.crs = srcEpsgCode
        self.dstEpsgCode = dstEpsgCode
        self.decimalSep = decimalSep

    def run(self):
        _rows = CSVLib.read(self.inputFile, self.fieldSep, self.decimalSep)

        rows = []
        for row in _rows:
            row['geometry'] = Point(row[self.xFieldName], row[self.yFieldName])
            rows.append(row)

        outputGdf = GeoDataFrame(rows, crs=self.crs)

        if not self.dstEpsgCode is None:
            outputGdf = outputGdf.to_crs(self.dstEpsgCode)
        return outputGdf



'''
Created on 15 janv. 2021

@author: tleduc

Copyright 2020 Thomas Leduc

This file is part of t4gpd.

t4gpd is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

t4gpd is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with t4gpd.  If not, see <https://www.gnu.org/licenses/>.
'''
from t4gpd.commons.IllegalArgumentTypeException import IllegalArgumentTypeException
from numpy import exp, log, log2, log10


class Entropy(object):
    '''
    classdocs
    '''

    def __init__(self, probs):
        '''
        Constructor
        '''
        for prob in probs:
            if not (0 <= prob <= 1):
                raise IllegalArgumentTypeException(prob, 'float value between 0.0 and 1.0')
        self.probs = probs

    @staticmethod
    def createFromIntValuesArray(intValues):
        '''
        # bincount([1]*5 + [3]*12 + [4]*33) -> array([ 0,  5,  0, 12, 33])
        _nvalues, _values = len(intValues), array(intValues)
        if (0 > min(_values)):
            _values = _values + min(_values)
        _probs = bincount(_values) / _nvalues
        return Entropy(_probs)
        '''
        mapOfNbOccurrences = dict()
        for value in intValues:
            if value in mapOfNbOccurrences:
                mapOfNbOccurrences[value] += 1
            else:
                mapOfNbOccurrences[value] = 1
        probs = [float(nbOccurr) / len(intValues) for nbOccurr in list(mapOfNbOccurrences.values())]
        return Entropy(probs)

    @staticmethod
    def createFromDoubleValuesArray(doubleValues, precision=1.0):
        intValues = [int(x / precision) for x in doubleValues]
        return Entropy.createFromIntValuesArray(intValues)

    @staticmethod
    def createFromString(string):
        arrayOfChars = [ord(c) for c in string]
        return Entropy.createFromIntValuesArray(arrayOfChars)

    def h(self, base=exp(1)):
        if (exp(1) == base):
            return -sum([prob * log(prob) for prob in self.probs if (0.0 < prob)])
        elif (2 == base):
            return -sum([prob * log2(prob) for prob in self.probs if (0.0 < prob)])
        elif (10 == base):
            return -sum([prob * log10(prob) for prob in self.probs if (0.0 < prob)])
        elif (0 < base):
            _tmp = 1 / log(base)
            return -sum([prob * log(prob) * _tmp for prob in self.probs if (0.0 < prob)])
        raise IllegalArgumentTypeException(base, 'must be strictly positive')



'''
Created on 10 juil. 2022

@author: tleduc

Copyright 2020-2022 Thomas Leduc

This file is part of t4gpd.

t4gpd is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

t4gpd is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with t4gpd.  If not, see <https://www.gnu.org/licenses/>.
'''

from qgis.PyQt.QtCore import QRectF
from qgis.PyQt.QtGui import QFont
from qgis.core import (QgsLayoutExporter, QgsLayoutItemLabel, QgsLayoutItemLegend,
                       QgsLayoutItemMap, QgsLayoutItemScaleBar,
                       QgsLayoutPoint, QgsLayoutSize, QgsLegendStyle,
                       QgsPrintLayout, QgsProject,
                       QgsUnitTypes)

from t4gpd.commons.IllegalArgumentTypeException import IllegalArgumentTypeException
from t4gpd.demos.GeoDataFrameDemos import GeoDataFrameDemos
from t4gpd.pyqgis.AddMemoryLayer import AddMemoryLayer
from t4gpd.pyqgis.SetSymbolLib import SetSymbolLib
from t4gpd.pyqgis.ZoomLib import ZoomLib


class MapPrinter(object):
    '''
    classdocs
    '''
    PDF = 1
    PNG = 2
    SVG = 3

    @staticmethod
    def __getLayout(layerExtent, xOffset=0, yOffset=0, legendXOffset=238, legendYOffset=1, scaleBarXOffset=60,
        scaleBarYOffset=120, scaleBarNSegm=3, scalebarBg=False, fileFormat='a4landscape',
        textLabel=None, labelXOffset=10, labelYOffset=10, fontLabel=QFont('Cambria', 34, QFont.Bold),
        numUnitsPerSegment=None, scaleBarFontSize=QFont('Cambria', 24), fontLegend=None, symbolHeight=None,
        legendBg=None):
        if 'a4landscape' == fileFormat:
            width, height = 297, 210
        elif 'a4portrait' == fileFormat:
            width, height = 210, 297
        elif 'square' == fileFormat:
            width, height = 210, 210
        layoutSize = QgsLayoutSize(width=width, height=height, units=QgsUnitTypes.LayoutMillimeters)

        project = QgsProject.instance()
        layout = QgsPrintLayout(project)
        layout.initializeDefaults()
        layout.pageCollection().page(0).setPageSize(layoutSize)

        # CANVAS
        myMap = QgsLayoutItemMap(layout)
        myMap.setRect(QRectF(layerExtent.xMinimum(), layerExtent.yMinimum(), layerExtent.xMaximum(), layerExtent.yMaximum()))
        myMap.setExtent(layerExtent)
        layout.addItem(myMap)
        myMap.attemptResize(layoutSize)

        # TITLE
        if textLabel is not None:
            label = QgsLayoutItemLabel(layout)
            label.setText(textLabel)
            label.setFont(fontLabel)
            label.adjustSizeToText()
            layout.addItem(label)
            label.attemptMove(QgsLayoutPoint(labelXOffset, labelYOffset, QgsUnitTypes.LayoutMillimeters))
            label.setBackgroundEnabled(1)

        # LEGEND
        legend = QgsLayoutItemLegend(layout)
        legend.setTitle('Legend')
        legend.setLinkedMap(myMap)
        if fontLegend is not None:
            legend.setStyleFont(QgsLegendStyle.Title, fontLegend)
            legend.setStyleFont(QgsLegendStyle.Subgroup, fontLegend)
            legend.setStyleFont(QgsLegendStyle.SymbolLabel, fontLegend)
        if symbolHeight is not None:
            legend.setSymbolHeight(symbolHeight)
        layout.addItem(legend)
        legend.attemptMove(QgsLayoutPoint(legendXOffset, legendYOffset, QgsUnitTypes.LayoutMillimeters))
        # ## legend.setFrameEnabled(True)
        if (legendBg is not None) and legendBg:
            legend.setBackgroundEnabled(1)
        # legend.setBackgroundEnabled(1 if scalebarBg else 0)

        # SCALE BAR
        scaleBar = QgsLayoutItemScaleBar(layout)
        # Style possibilities are: 'Single Box', 'Double Box', 'Line Ticks Middle', 'Line Ticks Down', 'Line Ticks Up', 'Numeric'
        scaleBar.setStyle('Single Box')
        scaleBar.setLinkedMap(myMap)
        scaleBar.applyDefaultSize()
        scaleBar.setNumberOfSegmentsLeft(0)
        scaleBar.setNumberOfSegments(scaleBarNSegm)
        if numUnitsPerSegment is not None:
            scaleBar.setUnitsPerSegment(numUnitsPerSegment)
        # scaleBar.setFont(scaleBarFontSize)
        layout.addItem(scaleBar)
        scaleBar.attemptMove(QgsLayoutPoint(scaleBarXOffset, scaleBarYOffset, QgsUnitTypes.LayoutMillimeters))
        # ## scaleBar.setFrameEnabled(True)
        scaleBar.setBackgroundEnabled(1 if scalebarBg else 0)

        return layout

    @staticmethod
    def __outputTypeController(outputFileName):
        tmp = outputFileName.lower()
        if tmp.endswith('.pdf'):
            return MapPrinter.PDF
        elif tmp.endswith('.png'):
            return MapPrinter.PNG
        elif tmp.endswith('.svg'):
            return MapPrinter.SVG
        raise IllegalArgumentTypeException(outputFileName, '.png, .pdf or .svg filename suffix')

    @staticmethod
    def print(outputFileName, layerExtent, xOffset=0, yOffset=0, legendXOffset=238, legendYOffset=1,
                  scaleBarXOffset=15, scaleBarYOffset=185, scaleBarNSegm=3, scalebarBg=False,
                  fileFormat='a4landscape', textLabel=None, labelXOffset=10, labelYOffset=10,
                  fontLabel=QFont('Cambria', 34, QFont.Bold), numUnitsPerSegment=None,
                  scaleBarFontSize=QFont('Cambria', 24), fontLegend=None, symbolHeight=None,
                  legendBg=None):
        layout = MapPrinter.__getLayout(layerExtent, xOffset, yOffset,
                                        legendXOffset, legendYOffset, scaleBarXOffset,
                                        scaleBarYOffset, scaleBarNSegm, scalebarBg,
                                        fileFormat, textLabel, labelXOffset, labelYOffset,
                                        fontLabel, numUnitsPerSegment, scaleBarFontSize,
                                        fontLegend, symbolHeight, legendBg)
        layoutExporter = QgsLayoutExporter(layout)
        tmp = MapPrinter.__outputTypeController(outputFileName)

        if (MapPrinter.PDF == tmp):
            layoutExporter.exportToPdf(outputFileName, QgsLayoutExporter.PdfExportSettings())
        elif (MapPrinter.PNG == tmp):
            layoutExporter.exportToImage(outputFileName, QgsLayoutExporter.ImageExportSettings())
        elif (MapPrinter.SVG == tmp):
            layoutExporter.exportToSvg(outputFileName, QgsLayoutExporter.SvgExportSettings())

    @staticmethod
    def test():
        buildings = GeoDataFrameDemos.districtGraslinInNantesBuildings()
        roi = GeoDataFrameDemos.coursCambronneInNantes()
        path = GeoDataFrameDemos.coursCambronneInNantesPath()
        trees = GeoDataFrameDemos.districtGraslinInNantesTrees()

        for ext in ['pdf', 'png', 'svg']:
            QgsProject.instance().clear()

            buildings_layer = AddMemoryLayer(buildings, 'buildings').run()
            roi_layer = AddMemoryLayer(roi, 'Region of Interest').run()
            path_layer = AddMemoryLayer(path, 'Path').run()
            trees_layer = AddMemoryLayer(trees, 'Trees').run()

            SetSymbolLib.setFillSymbol(buildings_layer, color='black', width_border='0')
            SetSymbolLib.setSimpleOutlineFillSymbol(roi_layer, color='red', width='1.1', penstyle='dot')
            SetSymbolLib.setLineSymbol(path_layer, color='black', penstyle='solid', width='0.55')
            SetSymbolLib.setMarkerSymbol(trees_layer, color='green', size='1.9', name='circle')

            ZoomLib.setExtent(roi)

            roiExtent = roi_layer.extent().buffered(10.0)

            MapPrinter.print(f'/tmp/test.{ext}', roiExtent, xOffset=0, yOffset=0, legendXOffset=155,
                legendYOffset=155, scaleBarXOffset=5, scaleBarYOffset=192,
                scaleBarNSegm=3, scalebarBg=True, fileFormat='square',
                textLabel='Map title', labelXOffset=5, labelYOffset=10,
                fontLabel=QFont('Arial', 24, QFont.Bold), numUnitsPerSegment=None,
                scaleBarFontSize=QFont('Arial', 18), fontLegend=QFont('Arial', 16),
                symbolHeight=6)

'''
# COPY-PASTE TO TEST IN THE PYQGIS CONSOLE:
from t4gpd.pyqgis.MapPrinter import MapPrinter
MapPrinter.test()
'''
